# coding=utf-8
# ======================================
# File:     finance.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-09-30
# Desc:
#   Back-test trade cost calculation and
#   Cash investment simulation functions.
# ======================================

import numpy as np
import pandas as pd
from numba import njit
from collections import Iterable

from .utilfuncs import ALL_COST_PARAMETERS


def set_cost(**kwargs):
    """ 新建一个cost字典

    交易成本的估算依赖四种类型的成本：
    1,  fixed_fee：固定费用，在交易过程中产生的固定现金费用，与交易金额和交易量无关，买入与卖出固定费用可以不同
        固定费用类型与固定费率或最低费用不能同时存在，当固定费用不为0时，直接使用固定费用作为交易费用，忽略其他参数
        交易费用 = 固定费用
    2,  rate：type：float，交易费率，交易过程中的固定费率，与交易金额成正比，买入与卖出的交易费率可以不同
        交易费用 = 交易金额 * 交易费率
    3,  min_fee: float，最低交易费用，当设置了交易费率时，按照交易费率计算费用，但如果按照交易费率算出的费用低于
        最低费用，则使用最低费用。例如：
            卖出100股股票，假设每股价格10元，交易费率为千分之一，最低费用为5元，则
            根据费率计算出交易费用为1元，但由于最低费用为5元，因而交易费用为5元
    4,  slipage：type：float，交易滑点，或者叫二阶费率。
        用于模拟交易过程中由于交易延迟或买卖冲击形成的交易成本，滑点绿表现为一个关于交易量的函数, 交易
        滑点成本等于该滑点率乘以交易金额： 滑点成本 = f(交易金额） * 交易成本

    Parameters
    ----------
    kwargs: dict
        可用的成本参数包括：
        - buy_fix: float, 固定买入费用，如果设置了固定费用，则费率和最低费用会被忽略
        - sell_fix: float, 固定卖出费用，如果设置了固定费用，则费率和最低费用会被忽略
        - buy_rate: float, 买入费率，实际买入费用等于买入金额乘以费率
        - sell_rate: float, 卖出费率，实际卖出费用等于卖出金额乘以费率
        - buy_min: float, 最低买入费用，当按费率计算的费用小于该费用时，按最低费用计算
        - sell_min: float, 最低卖出费用，当按费率计算的费用小于该费用时，按最低费用计算
        - slipage: float, 滑点，用于估算交易价格损失，金额为滑点乘以交易金额

    Returns
    -------
    cost : dict of float, cost parameters

    """
    cost = dict(buy_fix=0.0,
                sell_fix=0.0,
                buy_rate=0.003,
                sell_rate=0.001,
                buy_min=5.0,
                sell_min=0.0,
                slipage=0.0)

    update_cost(cost, **kwargs)
    return cost


def update_cost(c, **kwargs):
    """ 更新cost字典中各个值

    Parameters
    ----------
    kwargs: dict
        可用的成本参数包括：
        - buy_fix: float, 固定买入费用，如果设置了固定费用，则费率和最低费用会被忽略
        - sell_fix: float, 固定卖出费用，如果设置了固定费用，则费率和最低费用会被忽略
        - buy_rate: float, 买入费率，实际买入费用等于买入金额乘以费率
        - sell_rate: float, 卖出费率，实际卖出费用等于卖出金额乘以费率
        - buy_min: float, 最低买入费用，当按费率计算的费用小于该费用时，按最低费用计算
        - sell_min: float, 最低卖出费用，当按费率计算的费用小于该费用时，按最低费用计算
        - slipage: float, 滑点，用于估算交易价格损失，金额为滑点乘以交易金额

    Returns
    -------
    c: dict, 更新后的成本参数
    """
    if not isinstance(c, dict):
        raise TypeError(f'cost should be a dict, got {type(c)} instead')
    if any(k not in ALL_COST_PARAMETERS for k in c.keys()):
        invalid_keys = [k for k in c.keys() if k not in ALL_COST_PARAMETERS]
        raise KeyError(f'invalid keys ({invalid_keys}) found in cost dict!')
    for k, v in kwargs.items():
        if k not in ALL_COST_PARAMETERS:
            continue
        # TODO: 此处应该validate输入值
        c[k] = v

    return c


@njit
def calculate(trade_values: np.ndarray,
              is_buying,
              fixed_fees,
              buy_fix,
              sell_fix,
              buy_rate,
              sell_rate,
              buy_min,
              sell_min,
              slipage) -> float:
    """直接调用对象，计算交易费率或交易费用

    采用两种模式计算：
        当fixed_fees为True时，采用固定费用模式计算，返回值为包含滑点的交易成本列表，
        当fixed_fees为False时，采用固定费率模式计算，返回值为包含滑点的交易成本率列表

    Parameters
    ----------
    trade_values: ndarray:
        总交易金额清单
    is_buying: bool:
        当前是否计算买入费用或费率, 默认True
    fixed_fees: bool:
        当前是否采用固定费用模式计算, 默认False
    buy_fix: float:
        买入固定费用
    sell_fix: float:
        卖出固定费用
    buy_rate: float:
        买入费率
    sell_rate: float:
        卖出费率
    buy_min: float:
        买入最低费用
    sell_min: float:
        卖出最低费用
    slipage: float:
        滑点

    Returns
    -------
    np.ndarray:
    """
    if is_buying is None:
        is_buying = True
    if fixed_fees is None:
        fixed_fees = False

    if fixed_fees:  # 采用固定费用模式计算, 返回固定费用及滑点成本，返回的是费用而不是费率
        if is_buying:
            return buy_fix + slipage * trade_values ** 2
        else:
            return sell_fix + slipage * trade_values ** 2
    else:  # 采用固定费率模式计算
        if is_buying:
            if buy_min == 0.:
                return buy_rate + slipage * trade_values
            else:
                min_rate = buy_min / (trade_values - buy_min)
                return np.fmax(buy_rate, min_rate) + slipage * trade_values
        else:
            if sell_min == 0.:
                return sell_rate - slipage * trade_values
            else:
                min_rate = -sell_min / trade_values
                # 当trade_values中有0值时，将产生inf，且传递到caller后会导致问题，因此需要清零
                min_rate[np.isinf(min_rate)] = 0
                return np.fmax(sell_rate, min_rate) + slipage * trade_values


@njit
def get_selling_result(prices: np.ndarray,
                       a_to_sell: np.ndarray,
                       moq,
                       buy_fix,
                       sell_fix,
                       buy_rate,
                       sell_rate,
                       buy_min,
                       sell_min,
                       slipage):
    """计算出售投资产品的要素

    Parameters
    ----------
    prices: ndarray, 投资产品的价格
    a_to_sell: ndarray, 计划卖出数量，其形式为计划卖出的股票的数量，通常为负，且其绝对值通常小于等于可出售的数量
    moq: float, 卖出股票的最小交易单位
    buy_fix: float, 买入固定费用
    sell_fix: float, 卖出固定费用
    buy_rate: float, 买入费率
    sell_rate: float, 卖出费率
    buy_min: float, 买入最低费用
    sell_min: float, 卖出最低费用
    slipage: float, 滑点

    Returns
    -------
    tuple: (a_sold, cash_gained, fee)
     - a_sold: ndarray          实际出售的资产份额
     - cash_gained: ndarray     扣除手续费后获得的现金
     - fee: ndarray             扣除的手续费
    """
    if moq == 0:
        a_sold = a_to_sell
    else:
        a_sold = np.trunc(a_to_sell / moq) * moq
    sold_values = a_sold * prices
    if sell_fix == 0:  # 固定交易费用为0，按照交易费率模式计算
        rates = calculate(trade_values=sold_values,
                          is_buying=False,
                          fixed_fees=False,
                          buy_fix=buy_fix,
                          sell_fix=sell_fix,
                          buy_rate=buy_rate,
                          sell_rate=sell_rate,
                          buy_min=buy_min,
                          sell_min=sell_min,
                          slipage=slipage)
        cash_gained = (-1 * sold_values * (1 - rates))
        fees = -(sold_values * rates)
    else:  # 固定交易费用不为0时，按照固定费率收取费用——直接从交易获得的现金中扣除
        fixed_fees = calculate(trade_values=sold_values,
                               is_buying=False,
                               fixed_fees=True,
                               buy_fix=buy_fix,
                               sell_fix=sell_fix,
                               buy_rate=buy_rate,
                               sell_rate=sell_rate,
                               buy_min=buy_min,
                               sell_min=sell_min,
                               slipage=slipage)
        fees = np.where(a_sold, fixed_fees, 0)
        cash_gained = - sold_values - fees
    return a_sold, cash_gained, fees


@njit
def get_purchase_result(prices: np.ndarray,
                        cash_to_spend: np.ndarray,
                        moq,
                        buy_fix,
                        sell_fix,
                        buy_rate,
                        sell_rate,
                        buy_min,
                        sell_min,
                        slipage):
    """获得购买资产时的要素

    Parameters
    ----------
    prices: ndarray, 投资组合中每只股票的当前单价
    cash_to_spend: ndarray, 买入金额，可用于买入股票或资产的计划金额
    moq: float, 最小交易单位
    buy_fix: float, 买入固定费用
    sell_fix: float, 卖出固定费用
    buy_rate: float, 买入费率
    sell_rate: float, 卖出费率
    buy_min: float, 买入最低费用
    sell_min: float, 卖出最低费用
    slipage: float, 滑点

    Returns
    -------
    tuple: (a_to_purchase, cash_spent, fee)
    a_to_purchase: ndarray,  代表所有股票分别买入的份额或数量
    cash_spent: ndarray,     花费的总金额，包括购买成本在内
    fee: ndarray,            花费的费用，购买成本，包括佣金和滑点等投资成本
    """
    # buy_min = buy_min
    # buy_fix = buy_fix
    if buy_fix == 0.:
        # 固定费用为0，估算购买一定金额股票的交易费率，考虑最小费用，将绝对值小于buy_min的金额置0
        # （因为在"allow_sell_short"模式下，cash_to_spend可能会小于零，代表买入负持仓）
        cash_to_spend = np.where(np.abs(cash_to_spend) < buy_min, 0, cash_to_spend)
        rates = calculate(trade_values=cash_to_spend,
                          is_buying=True,
                          fixed_fees=False,
                          buy_fix=buy_fix,
                          sell_fix=sell_fix,
                          buy_rate=buy_rate,
                          sell_rate=sell_rate,
                          buy_min=buy_min,
                          sell_min=sell_min,
                          slipage=slipage)
        # 根据moq计算实际购买份额，当价格为0的时候买入份额为0
        if moq == 0:  # moq为0，实际买入份额与期望买入份额相同
            a_purchased = np.where(prices,
                                   cash_to_spend / (prices * (1 + rates)),
                                   0.)
        else:  # moq不为零，实际买入份额必须是moq的倍数，因此实际买入份额通常小于期望买入份额
            a_purchased = np.where(prices,
                                   np.trunc(cash_to_spend / (prices * moq * (1 + rates))) * moq,
                                   0.)
        # 根据交易量计算交易费用，考虑最低费用的因素，当费用低于最低费用时，使用最低费用
        fees = np.where(a_purchased, np.fmax(a_purchased * prices * rates, buy_min), 0.)
        purchased_values = a_purchased * prices + fees
        cash_spent = np.where(a_purchased, -1 * purchased_values, 0.)
    else:  # self.buy_fix
        # 固定费用不为0，按照固定费用模式计算费用，忽略费率并且忽略最小费用，将绝对值小于buy_fix的金额置0
        # （因为在"allow_sell_short"模式下，cash_to_spend可能会小于零，代表买入负持仓）
        cash_to_spend = np.where(np.abs(cash_to_spend) < buy_fix, 0, cash_to_spend)
        fixed_fees = calculate(trade_values=cash_to_spend,
                               is_buying=True,
                               fixed_fees=True,
                               buy_fix=buy_fix,
                               sell_fix=sell_fix,
                               buy_rate=buy_rate,
                               sell_rate=sell_rate,
                               buy_min=buy_min,
                               sell_min=sell_min,
                               slipage=slipage)
        if moq == 0.:
            a_purchased = np.fmax(np.where(prices,
                                           (cash_to_spend - fixed_fees) / prices,
                                           0.),
                                  0.)
        else:
            a_purchased = np.fmax(np.where(prices,
                                           np.trunc((cash_to_spend - fixed_fees) / (prices * moq)) * moq,
                                           0.),
                                  0.)
        cash_spent = np.where(a_purchased, -1 * a_purchased * prices - fixed_fees, 0.)
        fees = np.where(a_purchased, fixed_fees, 0.)
    return a_purchased, cash_spent, fees


class CashPlan:
    """ 现金计划类，在策略回测的过程中用来模拟固定日期的现金投资额

    投资计划对象包含一组带时间戳的投资金额数据，用于模拟在固定时间的现金投入，可以实现对一次性现金投入和资金定投的模拟

    Attributes
    ----------
    first_date: str, 第一笔投资的日期
    last_date: str, 最后一笔投资的日期
    periods: int, 投资的总期数
    investment_amounts: list, 投资的金额列表
    dates: list, 投资的日期列表
    amounts: list, 投资的金额列表
    total: float, 投资的总金额
    ir: float, 投资的年化收益率
    closing_value: float, 投资的最终价值
    opening_value: float, 投资的初始价值

    Methods
    -------
    info()
        打印投资计划的基本信息
    reset_dates(dates: [list, str])
        重置投资计划的日期
    plan()
        返回投资计划的时间序列数据
    to_dict(key: str = 'cash_plan')
        返回投资计划的字典数据

    Operators
    ---------
    __add__(other)
        投资计划的加法运算，两个投资计划相加得到投资计划的并集
    __mul__(other)
        投资计划的乘法运算，投资计划乘以一个整数得到倍增的投资计划
    __rmul__(other)
        投资计划的乘法运算，一个整数乘以投资计划得到延长的投资计划，一个浮点数乘以投资计划得到倍增的投资计划

    Examples
    --------
    >>> plan = CashPlan(dates='2020-01-01', amounts=10000)
    >>> plan.info()
    <class 'qteasy.finance.CashPlan'>
    Investment is one-off amount of ¥10,000.00 on 2020-01-01
    Interest rate: 0.00%, equivalent final value: ¥10,000.00:
                amount
    2020-01-01   10000

    >>> plan2 = CashPlan(dates=['2020-01-01', '2020-02-01'], amounts=[10000, 20000])
    >>> plan2.info()
    <class 'qteasy.finance.CashPlan'>
    Investment is two times of ¥10,000.00 on 2020-01-01 and ¥20,000.00 on 2020-02-01
    Interest rate: 0.00%, equivalent final value: ¥30,000.00:
                amount
    2020-01-01   10000
    2020-02-01   20000

    >>> plan
    CashPlan(['20200101'], [10000], 0.0)
    >>> plan2
    CashPlan(['20200101', '20200201'], [10000, 20000], 0.0)
    >>> plan + plan2
    CashPlan(['20200101', '20200201'], [20000, 20000], 0.0)

    >>> plan2 + 10000
    CashPlan(['20200101', '20200201'], [20000, 30000], 0.0)
    >>> plan2 * 2
    CashPlan(['20200101', '20200201'], [20000, 40000], 0.0)
    >>> 2 * plan2
    CashPlan(['20200101', '20200201', '20200304', '20200404'], [10000, 20000, 10000, 20000], 0.0)
    """

    def __init__(self, dates, amounts: (int, float, [int], [float]), interest_rate: float = 0.0):
        """ 初始化投资计划

        Parameters
        ----------
        dates: str, datetime, list of str or list of datetime
            投资的日期，可以是一个日期字符串，也可以是一个日期字符串列表
        amounts: int, float, list of int or float
            投资的金额，可以是一个数字，也可以是一个数字列表
        interest_rate: float, Default 0.0
            投资的年化收益率，用于计算投资的最终价值

        Examples
        --------
        >>> plan = CashPlan(dates='2020-01-01', amounts=10000)
        >>> plan
        CashPlan(['20200101'], [10000], 0.0)
        >>> plan = CashPlan(dates=['2020-01-01', '2020-02-01'], amounts=[10000, 20000])
        >>> plan
        CashPlan(['20200101', '20200201'], [10000, 20000], 0.0)
        """

        if isinstance(amounts, (int, float)):
            amounts = [amounts]
        assert isinstance(amounts, (list, np.ndarray)), \
            f'TypeError: amounts should be a list of numbers, got {type(amounts)} instead'
        if isinstance(amounts, list):
            assert all([isinstance(amount, (int, float, np.int64, np.float64)) for amount in amounts]), \
                f'TypeError: amount should be number format, got unresolved format in amounts!'
            assert all([amount > 0 for amount in amounts]), f'InputError: Investment amount should be larger than 0'
        assert isinstance(dates, Iterable), f"Expect Iterable input dates, got {type(dates)} instead!"

        if isinstance(dates, str):
            dates = dates.replace(' ', '')
            dates = dates.split(',')
        try:
            dates = list(map(pd.to_datetime, dates))
        except Exception as e:
            raise KeyError(f'{e}, some of the input strings can not be converted to date time format!')

        assert len(amounts) == len(dates), \
            f'InputError: number of amounts should be equal to that of dates, can\'t match {len(amounts)} amounts in' \
            f' to {len(dates)} days.'

        self._cash_plan = pd.DataFrame(amounts, index=dates, columns=['amount']).sort_index()
        assert isinstance(interest_rate, float), \
            f'TypeError, interest rate should be a float number, got {type(interest_rate)}'
        assert 0. <= interest_rate <= 1., \
            f'InputError, interest rate should be between 0 and 100%, got {interest_rate:.2%}'
        self._ir = interest_rate

    @property
    def first_day(self):
        """ 返回投资第一天的日期

        Returns
        -------
        pd.Timestamp

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'], amounts=[10000, 10000, 15000])
        >>> plan.first_day
        Timestamp('2020-01-01 00:00:00')
        """
        if self.dates:
            return self.dates[0]
        else:
            return

    @property
    def last_day(self):
        """ 返回投资期最后一天的日期

        Returns
        -------
        pd.Timestamp

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'], amounts=[10000, 10000, 15000])
        >>> plan.last_day
        Timestamp('2021-01-01 00:00:00')
        """
        if self.dates:
            return self.dates[-1]
        else:
            return

    @property
    def period(self):
        """ 返回第一次投资到最后一次投资之间的时长，单位为天

        Returns
        -------
        int

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan.period
        366
        """
        if self.dates:
            return (self.last_day - self.first_day).days
        else:
            return 0

    @property
    def investment_count(self):
        """ 返回在整个投资计划期间的投资次数

        Returns
        -------
        int

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan.investment_count
        2
        """
        return len(self.dates)

    @property
    def dates(self):
        """ 返回整个投资计划期间的所有投资日期，按从先到后排列

        Returns
        -------
        list of pd.Timestamp

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'], amounts=[10000, 10000, 15000])
        >>> plan.dates
        [Timestamp('2020-01-01 00:00:00'),
         Timestamp('2020-06-01 00:00:00'),
         Timestamp('2021-01-01 00:00:00')]
        """
        return list(self.plan.index)

    @property
    def amounts(self):
        """ 返回整个投资计划期间的所有投资额列表，按从先到后排列

        Returns
        -------
        list of float

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'], amounts=[10000, 10000, 15000])
        >>> plan.amounts
        [10000.0, 15000.0, 10000.0]
        """
        return list(self.plan.amount)

    @property
    def total(self):
        """ 返回整个投资计划期间的投资额总和，不考虑利率

        Returns
        -------
        float, 所有投资金额的简单相加

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'], amounts=[10000, 10000, 15000])
        >>> plan.total
        35000.0
        """
        return self.plan.amount.sum()

    @property
    def ir(self):
        """ 无风险利率，年化利率

        Returns
        -------
        float
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'],
        ... amounts=[10000, 10000, 15000],
        ... interest_rate=0.03)
        >>> plan
        CashPlan(['20200101', '20200601', '20210101'], [10000, 15000, 10000], 0.03)
        >>> plan.ir
        0.03
        """
        return self._ir

    @ir.setter
    def ir(self, ir: float):
        """ 设置无风险利率 """
        assert isinstance(ir, float), f'The interest rate should be a float number, not {type(ir)}'
        assert 0. < ir < 1., f'Interest rate should be between 0 and 1'
        self._ir = ir

    @property
    def closing_value(self):
        """ 计算所有投资额按照无风险利率到最后一笔投资当天的终值

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'],
        ... amounts=[10000, 10000, 15000],
        ... interest_rate=0.03)
        >>> plan.closing_value
        35563.06
        """
        if self.ir == 0:
            return self.total
        else:
            df = self.plan.copy()
            df['days'] = (df.index[-1] - df.index).days
            df['fv'] = df.amount * (1 + self.ir) ** (df.days / 365.)
            return np.round(df.fv.sum(), 2)

    @property
    def opening_value(self):
        """ 计算所有投资额按照无风险利率在第一个投资日的现值

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101', '20200601'],
        ... amounts=[10000, 10000, 15000],
        ... interest_rate=0.03)
        >>> plan.opening_value
        34524.44
        """
        if self.ir == 0:
            return self.total
        else:
            df = self.plan.copy()
            df['days'] = (df.index - df.index[0]).days
            df['pv'] = df.amount / (1 + self.ir) ** (df.days / 365.)
            return np.round(df.pv.sum(), 2)

    @property
    def plan(self):
        """ 返回整个投资区间的投资计划

        Returns
        -------
        pandas.DataFrame

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan.plan
                    amount
        2020-01-01   10000
        2021-01-01   10000
        """

        return self._cash_plan

    def reset_dates(self, dates):
        """ 重设投资日期，dates必须为一个可迭代的日期时间序列，数量与CashPlan的投资
            期数相同，且可转换为datetime对象

        Parameters
        ----------
        dates: iterables
            一个可迭代的时间日期序列

        Returns
        -------
        None

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan
        CashPlan(['20200101', '20210101'], [10000, 10000], 0.0)
        >>> plan.reset_dates(['20200101', '20200601'])
        >>> plan
        CashPlan(['20200101', '20200601'], [10000, 10000], 0.0)
        """
        try:
            idx = pd.Index(dates, dtype='datetime64[ns]')
            self.plan.index = idx
        except Exception as e:
            print(f'{e}, ')

    def to_dict(self, keys: [list, tuple, np.ndarray] = None):
        """ 返回整个投资区间的投资计划，形式为字典。默认key为日期，如果明确给出keys，则使用参数keys

        Parameters
        ----------
        keys: iterables
            用于替代投资计划日期的可迭代对象

        Returns
        -------
        dict

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan
        CashPlan(['20200101', '20210101'], [10000, 10000], 0.0)
        >>> plan.to_dict()
        {Timestamp('2020-01-01 00:00:00'): 10000,
         Timestamp('2021-01-01 00:00:00'): 10000}
        >>> plan.to_dict(['1st invest', '2nd invest'])
        {'1st invest': 10000, '2nd invest': 10000}
        """

        if keys is None:
            return dict(self.plan.amount)
        else:
            # assert isinstance(keys, list), f'TypeError, keys should be list, got {type(keys)} instead.'
            assert len(keys) == len(self.amounts), \
                f'InputError, count of elements in keys should be same as that of investment amounts, expect ' \
                f'{len(self.amounts)}, got {len(keys)}'
            return dict(zip(keys, self.amounts))

    def info(self):
        """ 打印投资计划的所有信息

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan.info()
        <class 'qteasy.finance.CashPlan'>
        Investment contains multiple entries
        Investment Period from 2020-01-01 to 2021-01-01, lasting 366 days
        Total investment count: 2 entries, total invested amount: ¥20,000.00
        Interest rate: 0.00%, equivalent final value: ¥20,000.00:
                    amount
        2020-01-01   10000
        2021-01-01   10000
        memory usage: 136 bytes
        """
        import sys
        print(f'\n{type(self)}')
        if self.investment_count > 1:
            print('Investment contains multiple entries')
            print(f'Investment Period from {self.first_day.date()} to {self.last_day.date()}, '
                  f'lasting {self.period} days')
            print(f'Total investment count: {self.investment_count} entries, total invested amount: ¥{self.total:,.2f}')
            print(f'Interest rate: {self.ir:.2%}, equivalent final value: ¥{self.closing_value:,.2f}:')
        else:
            print(f'Investment is one-off amount of ¥{self.total:,.2f} on {self.first_day.date()}')
            print(f'Interest rate: {self.ir:.2%}, equivalent final value: ¥{self.closing_value:,.2f}:')
        print(self.plan)
        print(f'memory usage: {sys.getsizeof(self.plan)} bytes\n')

    def __add__(self, other):
        """ 将CashPlan与另一个对象相加

        两个CashPlan对象相加，得到一个金额合并后的的CashPlan对象
        CashPlan对象与一个int或float对象相加，得到的新的CashPlan对象的每笔投资都增加int或float数额

        Parameters
        ----------
        other: (int, float, CashPlan): 另一个对象，根据对象类型不同行为不同

        Returns
        -------
        cashplan

        Examples
        --------
        >>> plan1 = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan2 = CashPlan(dates=['20200601', '20210101'], amounts=[10000, 10000])
        >>> plan1 + 10000
        CashPlan(['20200101', '20210101'], [20000, 20000], 0.0)
        >>> plan1 + plan2
        CashPlan(['20200101', '20200601', '20210101'], [10000.0, 10000.0, 20000.0], 0.0)
        """

        if isinstance(other, (int, float)):
            # 当other是一个数字时，在新的CashPlan的每个投资日期的投资额上加上other元
            min_amount = np.min(self.amounts)
            assert other > -min_amount, \
                f'ValueError, the amount will cause illegal invest value in plan, {min_amount} - {other}'
            new_amounts = np.array(self.amounts) + other
            return CashPlan(self.dates, new_amounts, self.ir)
        elif isinstance(other, CashPlan):
            plan1 = self._cash_plan
            plan2 = other._cash_plan
            index_combo = list(plan1.index)
            index_combo.extend(list(plan2.index))
            index_combo = list(set(index_combo))  # 新的CashPlan的投资日期集合是两个CashPlan的并集
            plan1 = plan1.reindex(index_combo).fillna(0)  # 填充Nan值避免相加后产生Nan值
            plan2 = plan2.reindex(index_combo).fillna(0)
            plan = (plan1 + plan2).sort_index()
            if self.ir == 0:
                new_ir = other.ir
            else:
                new_ir = self.ir
            return CashPlan(list(plan.index), list(plan.amount), new_ir)
        else:
            raise TypeError(f'Only CashPlan and numbers are supported, got {type(other)}')

    def __radd__(self, other):
        """ 另一个对象other + CashPlan的结果与CashPlan + other相同， 即：
            other + CashPlan == CashPlan + other

        Parameters
        ----------
        other: (int, float, CashPlan): 另一个对象，根据对象类型不同行为不同

        Returns
        -------
        CashPlan: 新的CashPlan对象

        Examples
        --------
        >>> plan1 = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan2 = CashPlan(dates=['20200601', '20210101'], amounts=[10000, 10000])
        >>> 10000 + plan1
        CashPlan(['20200101', '20210101'], [20000, 20000], 0.0)
        >>> plan1 + plan2
        CashPlan(['20200101', '20200601', '20210101'], [10000.0, 10000.0, 20000.0], 0.0)
        """
        return self.__add__(other)

    def __mul__(self, other):
        """CashPlan乘以一个int或float返回一个新的CashPlan，它的投资数量和投资日期与CashPlan对象相同，每次投资额增加int或float倍

        Parameters
        ----------
        other: (int, float), 倍增一个cashplan的倍数

        Returns
        -------
        CashPlan: 新的CashPlan对象

        Examples
        --------
        >>> plan1 = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan1 * 2
        CashPlan(['20200101', '20210101'], [20000, 20000], 0.0)
        >>> plan1 * 0.5
        CashPlan(['20200101', '20210101'], [5000.0, 5000.0], 0.0)
        """
        assert isinstance(other, (int, float))
        assert other >= 0
        new_dates = list(self.dates)
        new_amounts = list(np.array(self.amounts) * other)
        return CashPlan(new_dates, new_amounts, self.ir)

    def __rmul__(self, other):
        """ other 乘以一个CashPlan的结果是一个新的CashPlan，结果取决于other的类型：

        other 为 int时，新的CashPlan的投资次数重复int次，投资额不变，投资日期按照相同的频率顺延，如果CashPlan只有一个投资日期时
            频率为一年
        other 为 float时，other * CashPlan == CashPlan * other

        Parameters
        ----------
        other: (int, float), 倍增或延长一个cashplan的倍数或次数

        Returns
        -------
        CashPlan: 新的CashPlan对象

        Examples
        --------
        >>> plan1 = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> 2 * plan1
        CashPlan(['20200101', '20210101', '20220103', '20230104'], [10000, 10000, 10000, 10000], 0.0)
        >>> 2.5 * plan1
        CashPlan(['20200101', '20210101'], [25000.0, 25000.0], 0.0)
        """
        assert isinstance(other, (int, float))
        if isinstance(other, int):
            assert other > 1
            one_day = pd.Timedelta(1, 'd')
            one_month = pd.Timedelta(31, 'd')
            one_quarter = pd.Timedelta(93, 'd')
            one_year = pd.Timedelta(365, 'd')
            if self.investment_count == 1:  # 如果只有一次投资，则以一年为间隔
                new_dates = [self.first_day + one_year * i for i in range(other)]
                return CashPlan(new_dates, self.amounts * other, self.ir)
            else:  # 如果有两次或以上投资，则计算首末两次投资之间的间隔，以此为依据计算未来投资间隔
                if self.investment_count == 2:  # 当只有两次投资时，新增投资的间距与频率与两次投资的间隔相同
                    time_offset = pd.Timedelta(self.period * 2, 'd')
                else:  # 当投资次数多于两次时，整个投资作为一个单元，未来新增投资为重复每个单元的投资。单元的跨度可以为月、季度及年
                    if self.period <= 28:
                        time_offset = one_month
                    elif self.period <= 90:
                        time_offset = one_quarter
                    elif self.period <= 365:
                        time_offset = one_year
                    else:
                        time_offset = one_year * (self.period // 365 + 1)
                # 获取投资间隔后，循环生成所有的投资日期
                original_dates = self.dates
                new_dates = self.dates
                for i in range(other - 1):
                    new_dates.extend([date + time_offset * (i + 1) + one_day for date in original_dates])
                return CashPlan(new_dates, self.amounts * other, self.ir)

        else:  # 当other是浮点数时，返回CashPlan * other 的结果
            return self.__mul__(other)

    def __repr__(self):
        """ 打印cash plan

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan
        CashPlan(['20200101', '20210101'], [10000, 10000], 0.0)
        """
        dates = self.plan.index.strftime('%Y%m%d').to_list()
        return f'CashPlan({dates}, {self.amounts}, {self.ir})'

    def __str__(self):
        """ 打印cash plan

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> print(cp)
        CashPlan(['20200101', '20210101'], [10000, 10000], 0.0)
        """
        return self._cash_plan.__str__()

    def __getitem__(self, item):
        """ 获取cash plan中某一天的投资金额

        Parameters
        ----------
        item: (int, datetimelike), 用于索引cash plan的索引器

        Returns
        -------
        dict, cash plan中的一个部分

        Examples
        --------
        >>> plan = CashPlan(dates=['20200101', '20210101'], amounts=[10000, 10000])
        >>> plan[0]
        {'amount': 10000}
        >>> plan['20200101']
        {'amount': 10000}
        >>> plan['2020-01-01']
        {'amount': 10000}
        """
        if isinstance(item, int):
            return self.plan.iloc[item].to_dict()
        else:
            return self.plan.loc[item].to_dict()


# TODO: 实现多种方式的定投计划，可定制周期、频率、总金额、单次金额等简单功能，同时还应支持递增累进式定投、按照公式定投等高级功能
def distribute_investment(amount: float,
                          start: str,
                          end: str,
                          periods: int,
                          freq: str) -> CashPlan:
    """ 将投资额拆分成一系列定投金额，并生成一个CashPlan对象

    Parameters
    ----------
    amount:
    start:
    end:
    periods:
    freq:

    Returns
    -------
    cashplan:
    """
    raise NotImplementedError



# coding=utf-8
# ======================================
# File:     broker.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-04-09
# Desc:
#   class Broker for trader to submit
# trading orders and get trading results
# Broker classes are supposed to be
# inherited from Broker, representing
# different trading platforms or brokers
# ======================================

from queue import Queue
from abc import abstractmethod, ABCMeta
from concurrent.futures import ThreadPoolExecutor, as_completed

import pandas as pd
import numpy as np
import time

from qteasy import QT_CONFIG

CASH_DECIMAL_PLACES = QT_CONFIG['cash_decimal_places']
AMOUNT_DECIMAL_PLACES = QT_CONFIG['amount_decimal_places']


class Broker(object):
    """ Broker是交易所的抽象，它接受交易订单并返回交易结果

    BaseBroker定义了Broker的基本接口，所有的Broker都必须继承自BaseBroker，
    以实现不同交易所的接口

    Attributes:
    -----------
    order_queue: Queue
        交易订单队列，每个交易订单都是一个list，包含多个交易订单
    result_queue: Queue
        交易结果队列，每个交易结果都是一个list，包含多个交易结果
    status: str
        Broker的状态，可以是 'init', 'running', 'stopped', 'paused'
        - init: Broker刚刚创建，还没有开始运行
        - running: Broker正在运行, 接受交易订单并返回交易结果
        - stopped: Broker已经停止运行, 停止所有操作，未完成的订单将被丢弃，并退出主循环
        - paused: Broker暂停运行，未完成的订单将被暂停，可以接受交易订单，但不会返回交易结果

    Methods:
    --------
    run()
        Broker的主循环，从order_queue中获取交易订单并处理，获得交易结果并放入result_queue中
    get_result(order): abstract method
        交易所处理交易订单并获取交易结果,在子类中必须实现这个方法
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        self.broker_name = 'BaseBroker'
        self.user_name = ''
        self.password = ''
        self.token = ''
        self.status = 'init'  # init, running, stopped, paused
        self.debug = False

        self.order_queue = Queue()
        self.result_queue = Queue()
        self.broker_messages = Queue()

    def run(self):
        """ Broker的主循环，从order_queue中获取交易订单并处理，获得交易结果并放入result_queue中
        order_queue中的每一个交易订单都由get_result函数来处理并获取交易结果，get_result函数
        的执行过程是IO intensive的，因此需要使用ThreadPoolExecutor来并行处理交易订单
        """
        if self.debug:
            self.post_message(f'is running...')
        self.status = 'init'
        try:
            while self.status != 'stopped':
                time.sleep(0.05)
                # 如果Broker处于暂停状态，则不处理交易订单
                if self.status == 'paused':
                    continue
                # 如果order_queue为空，则不处理交易订单
                if self.order_queue.empty():
                    continue
                # 使用ThreadPoolExecutor并行调用get_result函数处理交易订单，将结果put到result_queue中
                with ThreadPoolExecutor(max_workers=10) as executor:
                    futures = []
                    while not self.order_queue.empty():
                        # TODO: 此处改进，只有order_queue中的订单被Broker读取后，才修改订单的状态为'submitted'，之前状态为'created'
                        order = self.order_queue.get()
                        if self.debug:
                            self.post_message(f'is running, '
                                              f'got order(ID) {order["order_id"]} from order queue'
                                              f'({self.order_queue.unfinished_tasks} orders left in the queue)...')
                        futures.append(executor.submit(self.get_result, order))
                        self.order_queue.task_done()
                    # 获取交易结果并将其放入result_queue中
                    for future in as_completed(futures):
                        results = future.result()
                        if self.debug:
                            self.post_message(f'is running, got result {results} '
                                              f'and put to result queue...')
                        for result in results:
                            self.result_queue.put(result)
            else:
                # 如果Broker正常退出，处理尚未完成的交易订单
                if self.debug:
                    self.post_message(f'is stopped, will process unfinished orders...')
                while not self.order_queue.empty():
                    order = self.order_queue.get()
                    # TODO: 如果Broker正常退出，应该将未完成的交易订单返回给用户，并提示用户后取消订单
                    self.result_queue.put(self.get_result(order))
                    self.order_queue.task_done()
        except KeyboardInterrupt:
            # 如果Broker被用户强制退出，处理尚未完成的交易订单
            if self.debug:
                self.post_message('is stopped by user, will stop broker and process unfinished orders...')
            self.status = 'stopped'
            while not self.order_queue.empty():
                order = self.order_queue.get()
                # TODO: 如果Broker被用户强制退出，应该将未完成的交易订单返回给用户，并提示用户后取消订单
                self.result_queue.put(self.get_result(order))
                self.order_queue.task_done()
        except Exception as e:
            # 如果Broker出现异常，处理尚未完成的交易订单
            if self.debug:
                self.post_message(f'is stopped by exception: {e}, will stop broker and process unfinished orders...')
                import traceback
                traceback.print_exc()
            self.status = 'stopped'
            raise e

    def post_message(self, message: str):
        """ 将消息放入消息队列
        """
        if self.debug:
            message = f'[DEBUG]-{message}'
        message = f'[{self.broker_name}]: {message}'
        self.broker_messages.put(message)

    def get_result(self, order):
        """ 交易所处理交易订单并获取交易结果

        Parameters:
        -----------
        order: dict 交易订单dict，详细信息如下:
            {'order_id': 订单ID,
             'pos_id': position ID,
             'direction',
             'order_type',
             'qty',
             'price',
             'submitted_time',
             'status'}

        Returns:
        --------
        raw_trade_result: dict, 初步交易结果dict,详细信息如下：
            {'order_id',
             'filled_qty',
             'price',
             'transaction_fee',
             'execution_time',
             'canceled_qty',
             'delivery_amount',
             'delivery_status'
             }
        """

        if self.debug:
            self.post_message(f'get_result():\nsubmit order components of order(ID) {order["order_id"]}:\n'
                              f'quantity:{order["qty"]}\norder_price={order["price"]}\n'
                              f'order_direction={order["direction"]}\n')
        trade_results = self.transaction_result(
                order_qty=order['qty'],
                order_price=order['price'],
                direction=order['direction'],
        )
        raw_trade_results = []
        for trade_result in trade_results:
            result_type, qty, filled_price, fee = trade_result

            if not isinstance(result_type, str):
                raise TypeError(f'result_type should be str, but got {type(result_type)}')
            if result_type not in ['filled', 'partial-filled', 'canceled']:
                raise ValueError(f'result_type should be one of ["filled", "canceled"], but got {result_type}')
            if not isinstance(qty, (int, float)):
                raise TypeError(f'qty should be int or float, but got {type(qty)}')
            if qty <= 0:
                raise ValueError(f'qty should be greater than 0, but got {qty}')
            if qty > order['qty']:
                raise ValueError(f'qty should be less than or equal to order["qty"], but got {qty}')
            if not isinstance(filled_price, (int, float)):
                raise TypeError(f'filled_price should be int or float, but got {type(filled_price)}')
            if filled_price < 0:
                raise ValueError(f'filled_price should be greater than 0, but got {filled_price}')
            if result_type == 'canceled' and filled_price != 0:
                raise ValueError(f'filled_price should be 0 when result_type is "canceled", but got {filled_price}')
            if not isinstance(fee, (int, float)):
                raise TypeError(f'fee should be int or float, but got {type(fee)}')
            if fee < 0:
                raise ValueError(f'fee should be greater than 0, but got {fee}')

            # 确认数据格式正确后，将数据圆整到合适的精度，并组装为raw_trade_result
            if self.debug:
                self.post_message(f'method: get_result(): got transaction result for order(ID) {order["order_id"]}\n'
                                  f'result_type={result_type}, \nqty={qty}, \n'
                                  f'filled_price={filled_price}, \nfee={fee}')
            # 圆整qty、filled_qty和fee
            qty = round(qty, AMOUNT_DECIMAL_PLACES)
            filled_price = round(filled_price, CASH_DECIMAL_PLACES)
            transaction_fee = round(fee, CASH_DECIMAL_PLACES)

            filled_qty = 0
            canceled_qty = 0
            if result_type in ['filled', 'partial-filled']:
                filled_qty = qty
            elif result_type == 'canceled':
                canceled_qty = qty
            else:
                raise ValueError(f'Unknown result_type: {result_type}, should be one of ["filled", "canceled"]')

            current_datetime = pd.to_datetime('today')
            raw_trade_result = {
                'order_id':        order['order_id'],
                'filled_qty':      filled_qty,
                'price':           filled_price,
                'transaction_fee': transaction_fee,
                'execution_time':  current_datetime.strftime('%Y-%m-%d %H:%M:%S'),
                'canceled_qty':    canceled_qty,
                'delivery_amount': 0,
                'delivery_status': 'ND',
            }
            if self.debug:
                self.post_message(f'method get_result(): created raw trade result for order(ID) {order["order_id"]}:\n'
                                  f'{raw_trade_result}')

            raw_trade_results.append(raw_trade_result)

        return raw_trade_results

    @abstractmethod
    def transaction_result(self, order_qty, order_price, direction):
        """ 交易所处理交易订单并获取交易结果, 抽象方法，需要由用户在子类中实现

        Parameters:
        -----------
        order_qty: float
            交易订单数量
        order_price: float
            交易订单价格
        direction: str
            交易订单方向，'buy' 或者 'sell'

        Returns:
        --------
        tuple of tuples: ((result_type, qty, price, fee), (...), ...)
            result_type: str 交易结果类型:
                'filled' - 成交,
                'partial_filled' - 部分成交,
                'canceled' - 取消
                'failed' - 失败
            qty: float
                成交/取消数量，这个数字应该小于等于order_qty，且大于等于0
            price: float
                成交价格, 如果是取消交易，价格为0或任意数字
            fee: float
                交易费用，交易费用应该大于等于0

        Notes:
        ------
        1. 如果交易所返回的交易结果是部分成交，返回部分成交的结果
        2. 如果交易所返回的交易结果是分批成交，返回所有成交的结果
        3. 如果交易所返回的交易结果是取消，返回取消的结果
        4. 如果交易所返回的交易结果是失败，返回失败的结果

        Examples:
        ---------
        >>> broker = Broker()
        >>> # 交易所返回完全成交结果
        >>> broker.transaction_result(100, 10, 'buy')
        (('filled', 100, 10, 5),)
        >>> # 交易所返回部分成交结果
        >>> broker.transaction_result(100, 10, 'buy')
        (('partial_filled', 50, 10, 5),)
        >>> # 交易所返回分批成交结果
        >>> broker.transaction_result(200, 10, 'buy')
        (('partial_filled', 100, 10, 5), ('filled', 100, 10, 5))
        >>> # 交易所返回取消结果
        >>> broker.transaction_result(100, 10, 'buy')
        (('canceled', 100, 0, 0),)
        >>> # 交易所返回失败结果
        >>> broker.transaction_result(100, 10, 'buy')
        (('failed', 0, 0, 0),)
        """
        pass


class SimpleBroker(Broker):
    """ SimpleBroker接到交易订单后，立即返回交易结果
    交易结果总是完全成交，根据moq调整交易批量，根据设定的交易费率计算交易费用，滑点是按照百分比计算的

    Parameters
    ----------
    """

    def __init__(self):
        super(SimpleBroker, self).__init__()
        self.broker_name = 'SimpleBroker'

    def transaction_result(self, order_qty, order_price, direction):
        """ 订单立即成交

        Returns:
        --------
        result_type: str
            交易结果类型，'filled' - 成交
        qty: float
            成交数量
        price: float
            成交价格
        fee: float
            交易费用
        """
        qty = order_qty
        price = order_price
        fee: float = 5.
        return 'filled', qty, price, fee


class RandomBroker(Broker):
    """ RandomBroker接到交易订单后，随机等待一段时间再返回交易结果。交易结果随机，包含完全成交、部分成交和取消交易

    交易费用根据交易方向和交易价格计算，滑点是按照百分比计算的，比如0.01表示1%

    Parameters
    ----------
    fee_rate_buy: float, default 0.0001
        买入交易费率，比如0.0001表示万分之一
    fee_rate_sell: float, default 0.0003
        卖出交易费率，比如0.0003表示万分之三
    moq: float, default 100
        最小交易数量，比如100表示最小交易数量为100股，如果交易数量不是100的整数倍，那么会被圆整到100的整数倍
    delay: float, default 5.0
        交易所处理订单的时间，单位是秒
    price_deviation: float, default 0.01
        实际交易价格和报价之间的差异，比如0.01表示1%
    probabilities: list of float, default [0.8, 0.15, 0.05]
        交易结果的概率，比如[0.8, 0.15, 0.05]表示交易结果为'filled'的概率为80%，为'partial-filled'的概率为15%，为'canceled'的概率为5%

    """

    def __init__(self, fee_rate_buy=0.0001, fee_rate_sell=0.0003, moq=100, delay=5.0, price_deviation=0.01,
                 probabilities=[0.8, 0.15, 0.05]):
        super(RandomBroker, self).__init__()
        self.broker_name = 'RandomBroker'
        self.fee_rate_buy = fee_rate_buy
        self.fee_rate_sell = fee_rate_sell
        self.moq = moq
        self.delay = delay
        self.price_deviation = price_deviation
        self.probabilities = probabilities

    def transaction_result(self, order_qty, order_price, direction):
        """ 订单随机成交

        Returns:
        --------
        result_type: str
            交易结果类型，'filled' - 成交, 'partial-filled' - 部分成交, 'canceled' - 取消
        qty: float
            成交/取消数量
        price: float
            成交价格, 如果是取消交易，价格为0或任意数字
        fee: float
            交易费用
        """
        from time import sleep
        from random import random

        remain_qty = order_qty
        order_results = []

        while remain_qty > 0.001:
            result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=self.probabilities)
            trade_delay = random() * self.delay  # 模拟交易所处理订单的时间
            price_deviation = random() * self.price_deviation  # 模拟成交价格与报价之间的差异

            sleep(trade_delay)

            if result_type in ['filled', 'partial-filled']:
                filled_proportion = 1
                if result_type == 'partial-filled':
                    filled_proportion = np.random.choice([0.25, 0.5, 0.75], p=[0.3, 0.5, 0.2])  # 模拟交易所的部分成交
                qty = remain_qty * filled_proportion
                if self.moq > 0:
                    qty = np.trunc(qty / self.moq) * self.moq
                    if (qty < self.moq) and (remain_qty > self.moq):
                        qty = self.moq # 如果成交数量小于moq，但是剩余数量大于moq，那么成交数量就是moq
                    elif (qty < self.moq) and (remain_qty <= self.moq):
                        qty = remain_qty # 如果成交数量小于moq，且剩余数量也小于moq，那么成交数量就是剩余数量
                        result_type = 'filled'
                if direction == 'buy':
                    order_price = order_price * (1 + price_deviation)
                    transaction_fee = qty * order_price * self.fee_rate_buy
                elif direction == 'sell':
                    order_price = order_price * (1 - price_deviation)
                    transaction_fee = qty * order_price * self.fee_rate_sell
                else:
                    raise RuntimeError(f'invalid direction: {direction}')

            else:  # result_type == 'canceled'
                transaction_fee = 0
                order_price = 0
                qty = remain_qty

            remain_qty -= qty
            order_results.append((result_type, qty, order_price, transaction_fee))

        return tuple(order_results)


class NotImplementedBroker(Broker):
    """ NotImplementedBroker raises NotImplementedError when __init__() is called
    """

    def __init__(self):
        super(NotImplementedBroker, self).__init__()
        raise NotImplementedError('NotImplementedBroker is not implemented yet')


ALL_BROKERS = {
    'simple': SimpleBroker,
    'random': RandomBroker,
    'manual': NotImplementedBroker,
    'simulator': NotImplementedBroker,
}


# coding=utf-8
# ======================================
# Package:  qteasy
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-11
# Desc:
#   QTEASY:
#   A fast and easy-to-use quant-investment
#   strategy research tool kit.
# ======================================

import os
import warnings

import tushare as ts
import numpy as np
import logging
from logging.handlers import TimedRotatingFileHandler

import qteasy.utilfuncs
from .core import run, set_config, get_configurations, get_config
from .core import info, is_ready, configure, configuration, save_config, load_config, reset_config
from .core import get_basic_info, get_stock_info, get_data_overview, refill_data_source
from .core import get_history_data, filter_stock_codes, filter_stocks
from .core import reconnect_ds, get_table_info, get_table_overview
from .history import HistoryPanel
from .history import dataframe_to_hp, stack_dataframes
from .qt_operator import Operator
from .strategy import BaseStrategy, RuleIterator, GeneralStg, FactorSorter
from .built_in import built_ins, built_in_list, built_in_strategies, get_built_in_strategy
from .visual import candle
from .finance import CashPlan, set_cost, update_cost
from .database import DataSource, find_history_data
from ._arg_validators import QT_CONFIG, ConfigDict


# 解析qteasy的本地安装路径
QT_ROOT_PATH = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))
QT_ROOT_PATH = os.path.join(QT_ROOT_PATH, 'qteasy/')

# 准备从本地配置文件中读取预先存储的qteasy配置
qt_local_configs = {}

QT_CONFIG_FILE_INTRO = '# qteasy configuration file\n' \
                       '# following configurations will be loaded when initialize qteasy\n\n' \
                       '# example:\n' \
                       '# local_data_source = database\n\n'
# 读取configurations文件内容到config_lines列表中，如果文件不存在，则创建一个空文本文件
try:
    with open(os.path.join(QT_ROOT_PATH, 'qteasy.cfg')) as f:
        config_lines = f.readlines()

except FileNotFoundError as e:
    warnings.warn(f'qteasy.cfg not found, a new configuration file is created.')
    with open(os.path.join(QT_ROOT_PATH, 'qteasy.cfg'), mode='w', encoding='utf-8') as f:
        intro = QT_CONFIG_FILE_INTRO
        f.write(intro)

    config_lines = []  # 本地配置文件行
except Exception as e:
    warnings.warn(f'Error reading configuration file, all configurations will fall back to default! \n{e}')
    config_lines = []

# 解析config_lines列表，依次读取所有存储的属性，所有属性存储的方式为：
# config_key = value
for line in config_lines:
    if line[0] == '#':  # 忽略注释行
        continue
    line = line.split('=')
    # TODO: 读取配置文件时，字符串是不需要引号的，但是如果字符串用引号引起来，说明强制要求字符串类型，需要解决
    if len(line) == 2:
        arg_name = line[0].strip()
        read_value = line[1].strip()
        if read_value == 'True':
            read_value = True
        elif read_value == 'False':
            read_value = False
        elif read_value == 'None':
            read_value = None
        elif qteasy.utilfuncs.is_integer_like(read_value):
            read_value = int(read_value)
        elif qteasy.utilfuncs.is_float_like(read_value):
            read_value = float(read_value)
        else:
            pass

        arg_value = read_value
        try:
            qt_local_configs[arg_name] = arg_value
        except Exception as e:
            warnings.warn(f'{e}, invalid parameter: {arg_name}')

# 读取tushare token，如果读取失败，抛出warning
try:
    TUSHARE_TOKEN = qt_local_configs['tushare_token']
    ts.set_token(TUSHARE_TOKEN)
except Exception as e:
    warnings.warn(f'Failed Loading tushare_token, configure it in qteasy.cfg:\n'
                  f'tushare_token = your_token\n'
                  f'for more information, please visit: https://tushare.pro/')

# 读取其他本地配置属性，更新QT_CONFIG, 允许用户自定义参数存在
configure(only_built_in_keys=False, **qt_local_configs)

# 连接默认的本地数据源
QT_DATA_SOURCE = DataSource(
        source_type=QT_CONFIG['local_data_source'],
        file_type=QT_CONFIG['local_data_file_type'],
        file_loc=QT_CONFIG['local_data_file_path'],
        host=QT_CONFIG['local_db_host'],
        port=QT_CONFIG['local_db_port'],
        user=QT_CONFIG['local_db_user'],
        password=QT_CONFIG['local_db_password'],
        db_name=QT_CONFIG['local_db_name']
)

# 初始化默认交易日历
QT_TRADE_CALENDAR = QT_DATA_SOURCE.read_table_data('trade_calendar')
if not QT_TRADE_CALENDAR.empty:
    QT_TRADE_CALENDAR = QT_TRADE_CALENDAR
else:
    QT_TRADE_CALENDAR = None
    warnings.warn(f'trade calendar is not loaded, some utility functions may not work '
                  f'properly, to download trade calendar, run \n"qt.refill_data_source(tables=\'trade_calendar\')"')

# 设置qteasy运行过程中忽略某些numpy计算错误报警
np.seterr(divide='ignore', invalid='ignore')

# 设置qteasy回测交易报告以及错误报告的存储路径
QT_SYS_LOG_PATH = os.path.join(QT_ROOT_PATH, QT_CONFIG['sys_log_file_path'])
QT_TRADE_LOG_PATH = os.path.join(QT_ROOT_PATH, QT_CONFIG['trade_log_file_path'])

# 设置系统日志以及交易日志的存储路径，如果路径不存在，则新建一个文件夹
os.makedirs(QT_SYS_LOG_PATH, exist_ok=True)
os.makedirs(QT_TRADE_LOG_PATH, exist_ok=True)
debug_handler = logging.handlers.TimedRotatingFileHandler(filename=os.path.join(QT_SYS_LOG_PATH, 'qteasy.log'),
                                                          backupCount=3, when='midnight')
error_handler = logging.StreamHandler()
debug_handler.setLevel(logging.DEBUG)
error_handler.setLevel(logging.ERROR)
formatter = logging.Formatter('[%(asctime)s]:%(levelname)s - %(module)s -: %(message)s')
debug_handler.setFormatter(formatter)
error_handler.setFormatter(formatter)
logger_core = logging.getLogger('core')
logger_core.addHandler(debug_handler)
logger_core.addHandler(error_handler)
logger_core.setLevel(logging.INFO)
logger_core.propagate = False

logger_core.info('qteasy loaded!')



# coding=utf-8
# ======================================
# File:     test_qt.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Unittest for general functions of
#   qteasy.
# ======================================
import unittest

import qteasy as qt
import numpy as np
from qteasy.tsfuncs import stock_basic

from qteasy.tafuncs import bbands
from qteasy.tafuncs import sma

from qteasy.strategy import BaseStrategy, RuleIterator, GeneralStg


class TestLSStrategy(RuleIterator):
    """用于test测试的简单多空蒙板生成策略。基于RuleIterator策略模版，将下列策略循环应用到所有股票上
        同时，针对不同股票策略参数可以不相同

    该策略有两个参数，N与Price
    如果给出的历史数据不包含参考数据时，策略逻辑如下：
     - 计算OHLC价格平均值的N日简单移动平均，判断：当移动平均价大于等于Price时，状态为看多，否则为看空
    如果给出参考数据时，策略逻辑变为：
     - 计算OHLC价格平均值的N日简单移动平均，判断：当移动平均价大于等于当日参考数据时，状态为看多，否则为看空
    如果给出交易结果数据时，策略逻辑变为：
     - 计算OHLC价格平均值的N日简单移动平均，判断：当移动平均价大于等于上次交易价时，状态为看多，否则为看空

    """

    def __init__(self):
        super().__init__(name='test_LS',
                         description='test long/short strategy',
                         par_count=2,
                         par_types='discr, conti',
                         par_range=([1, 20], [2, 20]),
                         strategy_data_types='close, open, high, low',
                         data_freq='d',
                         window_length=5)
        pass

    def realize(self, h, r=None, t=None, pars=None):
        if pars is not None:
            n, price = pars
        else:
            n, price = self.pars
        h = h.T
        avg = (h[0] + h[1] + h[2] + h[3]) / 4
        ma = sma(avg, n)
        if r is not None:
            # 处理参考数据生成信号并返回
            ref_price = r[-1, 0]  # 当天的参考数据，r[-1
            if ma[-1] < ref_price:
                return 0
            else:
                return 1

        if t is not None:
            # 处理交易结果数据生成信号并返回
            last_price = t[4]  # 获取最近的交易价格
            if np.isnan(last_price):
                return 1  # 生成第一次交易信号
            if ma[-1] < last_price:
                return 1
            else:
                return 0

        if ma[-1] < price:
            return 0
        else:
            return 1


class TestSelStrategy(GeneralStg):
    """用于Test测试的通用交易策略，基于GeneralStrategy策略生成

    策略没有参数，选股周期为5D
    在每个选股周期内，按以下逻辑选择股票并设定多空状态：
    当历史数据不含参考数据和交易结果数据时：
     - 计算：今日变化率 = (今收-昨收)/平均股价(HLC平均股价)，
     - 选择今日变化率最高的两支，设定投资比率50%，否则投资比例为0
    当给出参考数据时，按下面逻辑设定多空：
     - 计算：今日相对变化率 = (今收-昨收)/HLC平均股价/参考数据
     - 选择相对变化率最高的两只股票，设定投资比率为50%，否则为0
    当给出交易结果数据时，按下面逻辑设定多空：
     - 计算：交易价差变化率 = (今收-昨收)/上期交易价格
     - 选择交易价差变化率最高的两只股票，设定投资比率为50%，否则为0
    """

    def __init__(self):
        super().__init__(name='test_SEL',
                         description='test portfolio selection strategy',
                         par_count=0,
                         par_types='',
                         par_range=(),
                         strategy_data_types='high, low, close',
                         data_freq='d',
                         strategy_run_freq='10d',
                         window_length=5,
                         )
        pass

    def realize(self, h, r=None, t=None):
        avg = np.nanmean(h, axis=(1, 2))
        dif = (h[:, :, 2] - np.roll(h[:, :, 2], 1, 1))
        dif_no_nan = np.array([arr[~np.isnan(arr)][-1] for arr in dif])
        if r is not None:
            # calculate difper while r
            ref_price = np.nanmean(r[:, 0])
            difper = dif_no_nan / avg / ref_price
            large2 = difper.argsort()[1:]
            chosen = np.zeros_like(avg)
            chosen[large2] = 0.5
            return chosen

        if t is not None:
            # calculate difper while t
            last_price = t[:, 4]
            if np.all(np.isnan(last_price)):
                return np.ones_like(avg) * 0.333
            difper = dif_no_nan / last_price
            large2 = difper.argsort()[1:]
            chosen = np.zeros_like(avg)
            chosen[large2] = 0.5
            return chosen

        difper = dif_no_nan / avg
        large2 = difper.argsort()[1:]
        chosen = np.zeros_like(avg)
        chosen[large2] = 0.5
        return chosen


class Cross_SMA_PS(qt.RuleIterator):
    """自定义双均线择时策略策略，产生的信号类型为交易信号"""

    def __init__(self):
        """这个均线择时策略只有三个参数：
            - SMA 慢速均线，所选择的股票
            - FMA 快速均线
            - M   边界值

            策略的其他说明

        """
        """
        必须初始化的关键策略参数清单：

        """
        super().__init__(
                pars=(25, 100, 0.01),
                par_count=3,
                par_types=['discr', 'discr', 'conti'],
                par_range=[(10, 250), (10, 250), (0.0, 0.5)],
                name='CUSTOM ROLLING TIMING STRATEGY',
                description='Customized Rolling Timing Strategy for Testing',
                strategy_data_types='close',
                window_length=200,
        )

    # 策略的具体实现代码写在策略的realize()函数中
    # 这个函数固定接受两个参数： hist_price代表特定组合的历史数据， params代表具体的策略参数
    def realize(self, h, r=None, t=None, pars=None):
        """策略的具体实现代码：
        s：短均线计算日期；l：长均线计算日期；m：均线边界宽度
        """
        f, s, m = pars
        # 临时处理措施，在策略实现层对传入的数据切片，后续应该在策略实现层以外事先对数据切片，保证传入的数据符合data_types参数即可
        h = h.T
        # 计算长短均线的当前值和昨天的值
        sma = qt.tafuncs.sma
        s_ma = sma(h[0], s)
        f_ma = sma(h[0], f)

        s_today, s_last = s_ma[-1], s_ma[-2]
        f_today, f_last = f_ma[-1], f_ma[-2]

        # 计算慢均线的停止边界，当快均线在停止边界范围内时，平仓，不发出买卖信号
        s_ma_u = s_today * (1 + m)
        s_ma_l = s_today * (1 - m)

        # 根据观望模式在不同的点位产生交易信号
        if (f_last < s_ma_u) and (f_today > s_ma_u):  # 当快均线自下而上穿过上边界，开多仓
            return 1.
        elif (f_last > s_ma_u) and (f_today < s_ma_u):  # 当快均线自上而下穿过上边界，平多仓
            return -1.
        elif (f_last > s_ma_l) and (f_today < s_ma_l):  # 当快均线自上而下穿过下边界，开空仓
            return -1.
        elif (f_last < s_ma_l) and (f_today > s_ma_l):  # 当快均线自下而上穿过下边界，平空仓
            return 1.
        else:  # 其余情况不产生任何信号
            return 0.


class Cross_SMA_PT(qt.RuleIterator):
    """自定义双均线择时策略策略，产生的信号类型为持仓目标信号"""

    def __init__(self):
        """这个均线择时策略只有三个参数：
            - SMA 慢速均线，所选择的股票
            - FMA 快速均线
            - M   边界值

            策略的其他说明

        """
        super().__init__(
                pars=(25, 100, 0.01),
                par_count=3,
                par_types=['discr', 'discr', 'conti'],
                par_range=[(10, 250), (10, 250), (0.0, 0.5)],
                name='CUSTOM ROLLING TIMING STRATEGY',
                description='Customized Rolling Timing Strategy for Testing',
                strategy_data_types='close',
                window_length=200,
        )

    # 策略的具体实现代码写在策略的_realize()函数中
    # 这个函数固定接受两个参数： hist_price代表特定组合的历史数据， params代表具体的策略参数
    def realize(self, h, r=None, t=None, pars=None):
        """策略的具体实现代码：
        s：短均线计算日期；l：长均线计算日期；m：均线边界宽度；hesitate：均线跨越类型"""
        f, s, m = pars
        # 临时处理措施，在策略实现层对传入的数据切片，后续应该在策略实现层以外事先对数据切片，保证传入的数据符合data_types参数即可
        h = h.T
        # 计算长短均线的当前值
        sma = qt.tafuncs.sma
        s_ma = sma(h[0], s)[-1]
        f_ma = sma(h[0], f)[-1]

        # 计算慢均线的停止边界，当快均线在停止边界范围内时，平仓，不发出买卖信号
        s_ma_u = s_ma * (1 + m)
        s_ma_l = s_ma * (1 - m)

        # 根据观望模式在不同的点位产生交易信号
        if s_ma_u < f_ma:  # 当快均线在上边界以上时，持有多头仓位
            return 1
        elif s_ma_l <= f_ma <= s_ma_u:  # 当快均线在上下边界之间时，清空所有持仓
            return 0
        elif f_ma < s_ma_l:  # 当快均线在下边界以下时，持有空头仓位
            return -1
        else:  # 其余情况不产生任何信号
            return 0


class TestQT(unittest.TestCase):
    """对qteasy系统进行总体测试"""

    def setUp(self):
        # 准备测试所需数据，确保本地数据源有足够的数据

        self.op = qt.Operator(strategies=['dma', 'macd'])
        print('  START TO TEST QT GENERAL OPERATIONS\n'
              '=======================================')
        self.op.set_parameter('dma', opt_tag=1, par_range=[(10, 250), (10, 250), (10, 250)])
        self.op.set_parameter('macd', opt_tag=1, par_range=[(10, 250), (10, 250), (10, 250)])
        self.op.signal_type = 'pt'

        qt.configure(benchmark_asset='000300.SH',
                     mode=1,
                     benchmark_asset_type='IDX',
                     asset_pool='000300.SH',
                     asset_type='IDX',
                     opti_output_count=50,
                     invest_start='20070110',
                     trade_batch_size=0.,
                     sell_batch_size=0.,
                     parallel=True)

        timing_pars1 = (165, 191, 23)
        timing_pars2 = {'000100': (77, 118, 144),
                        '000200': (75, 128, 138),
                        '000300': (73, 120, 143)}
        timing_pars3 = (115, 197, 54)
        self.op.set_blender('pos_2_0(s0, s1)')
        self.op.set_parameter(stg_id='dma', pars=timing_pars1)
        self.op.set_parameter(stg_id='macd', pars=timing_pars3)

    def test_configure(self):
        """测试参数设置
            通过configure设置参数
            通过QR_CONFIG直接设置参数
            设置不存在的参数时报错
            设置不合法参数时报错
            参数设置后可以重用

        """
        config = qt.QT_CONFIG
        self.assertEqual(config.mode, 1)
        qt.configure(mode=2)
        self.assertEqual(config.mode, 2)
        self.assertEqual(qt.QT_CONFIG.mode, 2)
        self.assertEqual(config.benchmark_asset, '000300.SH')
        self.assertEqual(config.benchmark_asset_type, 'IDX')
        self.assertEqual(config.asset_pool, '000300.SH')
        self.assertEqual(config.invest_start, '20070110')
        # test temp config_key in run() that works only in run()
        qt.run(self.op,
               mode=1,
               asset_pool='000001.SZ',
               asset_type='E',
               invest_start='20100101',
               visual=False)
        self.assertEqual(config.mode, 2)
        self.assertEqual(qt.QT_CONFIG.mode, 2)
        self.assertEqual(config.benchmark_asset, '000300.SH')
        self.assertEqual(config.benchmark_asset_type, 'IDX')
        self.assertEqual(config.asset_pool, '000300.SH')
        self.assertEqual(config.invest_start, '20070110')

        config_copy = config.copy()
        qt.configure(config_copy,
                     mode=1,
                     benchmark_asset='000002.SZ',
                     benchmark_asset_type='E')
        self.assertEqual(config.mode, 2)
        self.assertEqual(config.benchmark_asset, '000300.SH')
        self.assertEqual(config.benchmark_asset_type, 'IDX')
        self.assertEqual(config.asset_pool, '000300.SH')
        self.assertEqual(config.invest_start, '20070110')
        self.assertEqual(config_copy.mode, 1)
        self.assertEqual(config_copy.benchmark_asset, '000002.SZ')
        self.assertEqual(config_copy.benchmark_asset_type, 'E')
        self.assertEqual(config_copy.asset_pool, '000300.SH')
        self.assertEqual(config_copy.invest_start, '20070110')

    def test_configuration(self):
        """ 测试CONFIG的显示"""
        print(f'configuration without argument\n')
        qt.configuration()
        print(f'configuration with level=1\n')
        qt.configuration(level=1)
        print(f'configuration with level2\n')
        qt.configuration(level=2)
        print(f'configuration with level3\n')
        qt.configuration(level=3)
        print(f'configuration with level4\n')
        qt.configuration(level=4)
        print(f'configuration with level=1, up_to=3\n')
        qt.configuration(level=1, up_to=3)
        print(f'configuration with info=True\n')
        qt.configuration(default=True)
        print(f'configuration with info=True, verbose=True\n')
        qt.configuration(default=True, verbose=True)

    def test_run_mode_0(self):
        """测试策略的实时信号生成模式"""
        op = qt.Operator(strategies=['stema'], op_type='stepwise')
        op.set_parameter('stema', pars=(6,))
        qt.QT_CONFIG.mode = 0
        # qt.run(op)
        # TODO: running qteasy in mode 0 will enter interactive shell, which is not testable

    def test_run_mode_1(self):
        """测试策略的回测模式,结果打印但不可视化"""
        qt.configure(mode=1,
                     trade_batch_size=1,
                     visual=False,
                     trade_log=True,
                     invest_cash_dates='20070604', )
        qt.run(self.op)

    def test_run_mode_1_visual(self):
        """测试策略的回测模式，结果可视化但不打印"""
        print(f'test plot with no buy-sell points and position indicators')
        qt.configuration(up_to=1, default=True)
        qt.run(self.op,
               mode=1,
               trade_batch_size=1,
               visual=True,
               trade_log=False,
               buy_sell_points=False,
               show_positions=False,
               invest_cash_dates='20070616')

        print(f'test plot with both buy-sell points and position indicators')
        qt.configuration(up_to=1, default=True)
        qt.run(self.op,
               mode=1,
               trade_batch_size=1,
               visual=True,
               trade_log=False,
               buy_sell_points=True,
               show_positions=True,
               invest_cash_dates='20070604')

    def test_run_mode_2_montecarlo(self):
        """测试策略的优化模式，使用蒙特卡洛寻优"""
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='single',
               test_type='single',
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=False,
               visual=False)
        print(f'strategy optimization in Montecarlo algorithm with parallel ON')
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='single',
               test_type='single',
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in Montecarlo with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='multiple',
               test_type='single',
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in Montecarlo with multiple sub-idx_range testing')
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='multiple',
               test_type='multiple',
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)

    def test_run_mode_2_montecarlo_visual(self):
        """测试策略的优化模式，使用蒙特卡洛寻优"""
        print(f'strategy optimization in Montecarlo algorithm with parallel ON')
        qt.configuration(up_to=1, default=True)
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='single',
               test_type='single',
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20140601',
               opti_cash_dates='20060407',
               test_start='20120604',
               test_end='20201130',
               test_cash_dates='20140604',
               test_indicators='years,fv,return,mdd,v,ref,alpha,beta,sharp,info',
               # 'years,fv,return,mdd,v,ref,alpha,beta,sharp,info'
               indicator_plot_type='violin',
               parallel=True,
               visual=True)
        qt.configuration(up_to=1, default=True)

    def test_run_mode_2_grid(self):
        """测试策略的优化模式，使用网格寻优"""
        print(f'strategy optimization in grid search algorithm with parallel OFF')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='single',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=False,
               visual=False)
        print(f'strategy optimization in grid search algorithm with parallel ON')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='single',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in grid search with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='multiple',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in grid search with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='multiple',
               test_type='multiple',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)

    def test_run_mode_2_grid_visual(self):
        """测试策略的优化模式，使用网格寻优"""
        print(f'strategy optimization in grid search algorithm with parallel OFF')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='single',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=False,
               visual=True,
               indicator_plot_type=0)
        print(f'strategy optimization in grid search algorithm with parallel ON')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='single',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True,
               indicator_plot_type=1)
        print(f'strategy optimization in grid search with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='multiple',
               test_type='single',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True,
               indicator_plot_type=2)
        print(f'strategy optimization in grid search with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=0,
               opti_type='multiple',
               test_type='multiple',
               opti_grid_size=128,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True,
               indicator_plot_type=3)

    def test_run_mode_2_incremental(self):
        """测试策略的优化模式，使用递进步长蒙特卡洛寻优"""
        print(f'strategy optimization in incremental algorithm with parallel OFF')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=10,
               opti_min_volume=5E7,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=False,
               visual=False)
        print(f'strategy optimization in incremental algorithm with parallel ON')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in incremental with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_type='multiple',
               test_type='single',
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in incremental with multiple sub-idx_range testing')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_type='multiple',
               test_type='multiple',
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)

    def test_run_mode_2_incremental_visual(self):
        """测试策略的优化模式，使用递进步长蒙特卡洛寻优，结果以图表输出"""
        print(f'strategy optimization in incremental algorithm with parallel ON')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True)
        print(f'strategy optimization in incremental with multiple sub-idx_range optimization')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_type='multiple',
               test_type='single',
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True)

    def test_run_mode_2_predict(self):
        """测试策略的优化模式，使用蒙特卡洛预测方法评价优化结果"""
        print(f'strategy optimization in montecarlo algorithm with predictive montecarlo test')
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='single',
               test_type='montecarlo',
               opti_output_count=20,
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)
        print(f'strategy optimization in incremental with with predictive montecarlo test')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_type='single',
               test_type='montecarlo',
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=False)

    def test_run_mode_2_predict_visual(self):
        """测试策略的优化模式，使用蒙特卡洛预测方法评价优化结果，结果以图表方式输出"""
        print(f'strategy optimization in montecarlo algorithm with predictive montecarlo test')
        qt.run(self.op,
               mode=2,
               opti_method=1,
               opti_type='single',
               test_type='montecarlo',
               opti_output_count=20,
               opti_sample_count=200,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True)
        print(f'strategy optimization in incremental with with predictive montecarlo test')
        qt.run(self.op,
               mode=2,
               opti_method=2,
               opti_type='single',
               test_type='montecarlo',
               opti_r_sample_count=100,
               opti_reduce_ratio=0.2,
               opti_output_count=20,
               opti_max_rounds=50,
               opti_min_volume=5E9,
               opti_start='20060404',
               opti_end='20141231',
               test_start='20120604',
               test_end='20201130',
               parallel=True,
               visual=True)

    def test_built_in_timing(self):
        """测试内置的择时策略"""
        # 使用以下参数测试所有qt.built_in中的交易策略
        #   mode=1,
        #   asset_pool='000300.SH, 399006.SZ',
        #   start='20200101',
        #   end='20211231',
        #   trade_log=False,
        #   visual=False,
        # 其他均为默认参数
        print(f'testing built-in strategies')
        for strategy in qt.built_in_strategies():
            print(f'testing strategy {strategy}')
            op = qt.Operator(strategies=[strategy])
            qt.run(
                    op,
                    mode=1,
                    asset_pool='000300.SH, 399006.SZ',
                    invest_start='20200101',
                    invest_end='20211231',
                    trade_log=False,
                    visual=False,
            )

    def test_multi_share_mode_1(self):
        """test built-in strategy selecting finance
        """
        op = qt.Operator(strategies=['long', 'finance', 'signal_none'])
        all_shares = stock_basic()
        shares_banking = qt.filter_stock_codes(date='20070101', industry='银行')
        print('extracted banking share pool:')
        print(all_shares.loc[all_shares['ts_code'].isin(shares_banking)])
        shares_estate = list((all_shares.loc[all_shares.industry == "全国地产"]['ts_code']).values)
        qt.configure(asset_pool=shares_banking[0:10],
                     asset_type='E',
                     benchmark_asset='000300.SH',
                     benchmark_asset_type='IDX',
                     opti_output_count=50,
                     invest_start='20070101',
                     invest_end='20181231',
                     invest_cash_dates=None,
                     trade_batch_size=1.,
                     mode=1,
                     trade_log=True)
        op.set_parameter('long', pars=())
        op.set_parameter('finance', pars=(True, 'proportion', 'greater', 0, 0, 0.4),
                         strategy_run_freq='Q',
                         strategy_data_types='pe',
                         sort_ascending=True,
                         weighting='proportion',
                         condition='greater',
                         ubound=0,
                         lbound=0,
                         max_sel_count=0.4)
        op.set_parameter('signal_none', pars=())
        op.set_blender('avg(s0, s1, s2)', 'ls')
        op.info()
        print(f'test portfolio selecting from shares_estate: \n{shares_estate}')
        qt.configuration()
        qt.run(op, visual=True, trade_log=True, trade_batch_size=100)

    def test_many_share_mode_1(self):
        """test built-in strategy selecting finance
        """
        print(f'test portfolio selection from large quantities of shares')
        op = qt.Operator(strategies=['long', 'finance', 'signal_none'])
        qt.configure(asset_pool=qt.filter_stock_codes(date='20070101',
                                                      industry=['银行', '全国地产', '互联网', '环境保护', '区域地产',
                                                                '酒店餐饮', '运输设备', '综合类', '建筑工程', '玻璃',
                                                                '家用电器', '文教休闲', '其他商业', '元器件', 'IT设备',
                                                                '其他建材', '汽车服务', '火力发电', '医药商业', '汽车配件',
                                                                '广告包装', '轻工机械', '新型电力', '多元金融', '饲料',
                                                                '铜', '普钢', '航空', '特种钢',
                                                                '种植业', '出版业', '焦炭加工', '啤酒', '公路', '超市连锁',
                                                                '钢加工', '渔业', '农用机械', '软饮料', '化工机械', '塑料',
                                                                '红黄酒', '橡胶', '家居用品', '摩托车', '电器仪表', '服饰',
                                                                '仓储物流', '纺织机械', '电器连锁', '装修装饰', '半导体',
                                                                '电信运营', '石油开采', '乳制品', '商品城', '公共交通',
                                                                '陶瓷', '船舶'],
                                                      area=['深圳', '北京', '吉林', '江苏', '辽宁', '广东',
                                                            '安徽', '四川', '浙江', '湖南', '河北', '新疆',
                                                            '山东', '河南', '山西', '江西', '青海', '湖北',
                                                            '内蒙', '海南', '重庆', '陕西', '福建', '广西',
                                                            '上海']),
                     asset_type='E',
                     benchmark_asset='000300.SH',
                     benchmark_asset_type='IDX',
                     opti_output_count=50,
                     invest_start='20070101',
                     invest_end='20171228',
                     invest_cash_dates=None,
                     trade_batch_size=1.,
                     mode=1,
                     trade_log=False,
                     hist_dnld_parallel=0)
        print(f'in total a number of {len(qt.QT_CONFIG.asset_pool)} shares are selected!')
        op.set_parameter('long', pars=())
        op.set_parameter('finance', pars=(True, 'proportion', 'greater', 0, 0, 30),
                         strategy_run_freq='Q',
                         strategy_data_types='basic_eps',
                         sort_ascending=True,
                         weighting='proportion',
                         condition='greater',
                         ubound=0,
                         lbound=0,
                         max_sel_count=30)
        op.set_parameter('signal_none', pars=())
        op.set_blender('avg(s0, s1, s2)', 'close')
        qt.run(op, visual=False, trade_log=True)

    def test_op_stepwise(self):
        """测试stepwise模式下的operator的表，使用两个测试专用交易策略"""
        # confirm that operator running results are same in stepwise and batch type
        op_batch = qt.Operator(strategies=['dma', 'macd'], signal_type='pt', op_type='batch')
        op_stepwise = qt.Operator(strategies=['dma', 'macd'], signal_type='pt', op_type='step')
        for op in [op_batch, op_stepwise]:
            op.set_parameter(0, window_length=100, pars=(12, 26, 9))
            op.set_parameter(1, window_length=100, pars=(12, 26, 9))

        qt.configure(
                benchmark_asset='000300.SH',
                benchmark_asset_type='IDX',
                asset_pool='601398.SH, 600000.SH, 000002.SZ',
                asset_type='E',
                opti_output_count=50,
                invest_start='20190101',
                invest_end='20190331',
                trade_batch_size=1.,
                sell_batch_size=1.,
                parallel=True,
                trade_log=False
        )
        print('backtest in batch mode:')
        res_batch = op_batch.run(mode=1)
        print('backtest in stepwise mode:')
        res_stepwise = op_stepwise.run(mode=1)
        val_batch = res_batch["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values
        val_stepwise = res_stepwise["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values
        print(f'the result of batched operation is\n'
              f'shape: {val_batch.shape}\n'
              f'{val_batch}\n'
              f'and the result of stepwise operation is\n'
              f'shape: {val_stepwise.shape}\n'
              f'{val_stepwise}')

        self.assertTrue(np.allclose(val_batch, val_stepwise))
        self.assertEqual(res_batch['final_value'],
                         res_stepwise['final_value'])

        print('backtest in batch mode:')
        res_batch = op_batch.run(
                mode=1,
                invest_start='20180101',
                invest_end='20191231'
        )
        print('backtest in stepwise mode:')
        res_stepwise = op_stepwise.run(
                mode=1,
                invest_start='20180101',
                invest_end='20191231'
        )
        val_batch = res_batch["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values
        val_stepwise = res_stepwise["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values

        self.assertTrue(np.allclose(val_batch, val_stepwise))
        self.assertEqual(res_batch['final_value'],
                         res_stepwise['final_value'])

        # test operator that utilizes trade data
        stg1 = TestLSStrategy()
        stg2 = TestSelStrategy()
        stg1.window_length = 100
        stg2.window_length = 100
        stg2.strategy_run_freq = '2w'
        op_batch = qt.Operator(strategies=[stg1, stg2], signal_type='pt', op_type='batch')
        op_stepwise = qt.Operator(strategies=[stg1, stg2], signal_type='pt', op_type='stepwise')
        par_stg1 = {'000100': (20, 10),
                    '000200': (20, 10),
                    '000300': (20, 6)}
        par_stg2 = ()
        for op in [op_batch, op_stepwise]:
            op.set_parameter(0, pars=par_stg1, opt_tag=1, par_range=([1, 20], [2, 100]))
            op.set_parameter(1, pars=par_stg2, opt_tag=1)

        qt.configure(
                benchmark_asset='000300.SH',
                benchmark_asset_type='IDX',
                asset_pool='601398.SH, 600000.SH, 000002.SZ',
                asset_type='E',
                opti_output_count=50,
                invest_start='20190101',
                invest_end='20190331',
                opti_start='20190101',
                opti_end='20191231',
                test_start='20200101',
                test_end='20200331',
                trade_batch_size=100.,
                sell_batch_size=100.,
                parallel=True,
                trade_log=False
        )
        print('output result back testing with test data')

        print('backtest in batch mode:')
        res_batch = op_batch.run(mode=1)
        print('backtest in stepwise mode:')
        res_stepwise = op_stepwise.run(mode=1)
        val_batch = res_batch["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values
        val_stepwise = res_stepwise["complete_values"][["601398.SH", "600000.SH", "000002.SZ"]].values
        print(f'the result of batched operation is\n'
              f'{val_batch}\n'
              f'and the result of stepwise operation is\n'
              f'{val_stepwise}')

        print('backtest in batch mode in optimization mode:')
        op_batch.run(mode=2)
        print('backtest in stepwise mode in optimization mode')
        op_stepwise.run(mode=2)

        print('test stepwise mode with different sample freq')

    def test_sell_short(self):
        """ 测试sell_short模式是否能正常工作（买入卖出负份额）"""
        op = qt.Operator([Cross_SMA_PS()], signal_type='PS')
        op.set_parameter(0, pars=(23, 100, 0.02))
        res = qt.run(op,
                     mode=1,
                     invest_start='20060101',
                     allow_sell_short=False,
                     trade_log=True,
                     visual=True)
        no_short_in_res = np.all(res['oper_count'].short == 0)
        self.assertTrue(no_short_in_res)
        res = qt.run(op,
                     mode=1,
                     invest_start='20060101',
                     allow_sell_short=True,
                     trade_log=True,
                     visual=True)
        no_short_in_res = np.all(res['oper_count'].short == 0)
        self.assertFalse(no_short_in_res)
        op = qt.Operator([Cross_SMA_PT()], signal_type='PT')
        op.set_parameter(0, (23, 100, 0.02))
        res = qt.run(op, mode=1,
                     invest_start='20060101',
                     allow_sell_short=False,
                     trade_log=True,
                     visual=True)
        no_short_in_res = np.all(res['oper_count'].short == 0)
        self.assertTrue(no_short_in_res)
        res = qt.run(op, mode=1,
                     invest_start='20060101',
                     allow_sell_short=True,
                     trade_log=True,
                     visual=True)
        no_short_in_res = np.all(res['oper_count'].short == 0)
        self.assertFalse(no_short_in_res)


if __name__ == '__main__':
    # get all stock prices from year 2020 to year 2022
    qt.refill_data_source(qt.QT_DATA_SOURCE, tables='stock_daily', start_date='20200101', end_date='20221231')

    # get index prices of 000300 and 399006 data from 2005 to year 2022
    qt.refill_data_source(qt.QT_DATA_SOURCE, tables='index_daily', start_date='20050101', end_date='20221231',
                          code_range='000300,399006')

    # get hourly price data for a few stocks in year 2016
    qt.refill_data_source(qt.QT_DATA_SOURCE, tables='stock_hourly', start_date='20160101', end_date='20161231',
                          code_range=['000001', '000002', '000005', '000006', '000007', '000918', '000819',
                                      '000899'])
    unittest.main()


# coding=utf-8
# ======================================
# File:     test_core_sub_funcs.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Unittest for all qteasy core
#   sub-functions.
# ======================================
import unittest

import qteasy as qt
import pandas as pd
import numpy as np

from qteasy import QT_DATA_SOURCE

from qteasy.space import Space, space_around_centre


class TestCoreSubFuncs(unittest.TestCase):
    """Test all functions in core.py"""

    def setUp(self):
        pass

    def test_input_to_list(self):
        print('Testing input_to_list() function')
        input_str = 'first'
        self.assertEqual(qt.utilfuncs.input_to_list(input_str, 3), ['first', 'first', 'first'])
        self.assertEqual(qt.utilfuncs.input_to_list(input_str, 4), ['first', 'first', 'first', 'first'])
        self.assertEqual(qt.utilfuncs.input_to_list(input_str, 2, None), ['first', 'first'])
        input_list = ['first', 'second']
        self.assertEqual(qt.utilfuncs.input_to_list(input_list, 3), ['first', 'second', None])
        self.assertEqual(qt.utilfuncs.input_to_list(input_list, 4, 'padder'), ['first', 'second', 'padder', 'padder'])
        self.assertEqual(qt.utilfuncs.input_to_list(input_list, 1), ['first', 'second'])
        self.assertEqual(qt.utilfuncs.input_to_list(input_list, -5), ['first', 'second'])

    def test_point_in_space(self):
        sp = Space([(0., 10.), (0., 10.), (0., 10.)])
        p1 = (5.5, 3.2, 7)
        p2 = (-1, 3, 10)
        self.assertTrue(p1 in sp)
        print(f'point {p1} is in space {sp}')
        self.assertFalse(p2 in sp)
        print(f'point {p2} is not in space {sp}')
        sp = Space([(0., 10.), (0., 10.), range(40, 3, -2)], 'conti, conti, enum')
        p1 = (5.5, 3.2, 8)
        self.assertTrue(p1 in sp)
        print(f'point {p1} is in space {sp}')

    def test_space_in_space(self):
        print('test if a space is in another space')
        sp = Space([(0., 10.), (0., 10.), (0., 10.)])
        sp2 = Space([(0., 10.), (0., 10.), (0., 10.)])
        self.assertTrue(sp2 in sp)
        self.assertTrue(sp in sp2)
        print(f'space {sp2} is in space {sp}\n'
              f'and space {sp} is in space {sp2}\n'
              f'they are equal to each other\n')
        sp2 = Space([(0, 5.), (2, 7.), (3., 9.)])
        self.assertTrue(sp2 in sp)
        self.assertFalse(sp in sp2)
        print(f'space {sp2} is in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'{sp2} is a sub space of {sp}\n')
        sp2 = Space([(0, 5), (2, 7), (3., 9)])
        self.assertFalse(sp2 in sp)
        self.assertFalse(sp in sp2)
        print(f'space {sp2} is not in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'they have different types of axes\n')
        sp = Space([(0., 10.), (0., 10.), range(40, 3, -2)])
        self.assertFalse(sp in sp2)
        self.assertFalse(sp2 in sp)
        print(f'space {sp2} is not in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'they have different types of axes\n')

    def test_space_around_centre(self):
        sp = Space([(0., 10.), (0., 10.), (0., 10.)])
        p1 = (5.5, 3.2, 7)
        ssp = space_around_centre(space=sp, centre=p1, radius=1.2)
        print(ssp.boes)
        print('\ntest multiple diameters:')
        self.assertEqual(ssp.boes, [(4.3, 6.7), (2.0, 4.4), (5.8, 8.2)])
        ssp = space_around_centre(space=sp, centre=p1, radius=[1, 2, 1])
        print(ssp.boes)
        self.assertEqual(ssp.boes, [(4.5, 6.5), (1.2000000000000002, 5.2), (6.0, 8.0)])
        print('\ntest points on edge:')
        p2 = (5.5, 3.2, 10)
        ssp = space_around_centre(space=sp, centre=p1, radius=3.9)
        print(ssp.boes)
        self.assertEqual(ssp.boes, [(1.6, 9.4), (0.0, 7.1), (3.1, 10.0)])
        print('\ntest enum spaces')
        sp = Space([(0, 100), range(40, 3, -2)], 'discr, enum')
        p1 = [34, 12]
        ssp = space_around_centre(space=sp, centre=p1, radius=5, ignore_enums=False)
        self.assertEqual(ssp.boes, [(29, 39), (22, 20, 18, 16, 14, 12, 10, 8, 6, 4)])
        print(ssp.boes)
        print('\ntest enum space and ignore enum axis')
        ssp = space_around_centre(space=sp, centre=p1, radius=5)
        self.assertEqual(ssp.boes, [(29, 39),
                                    (40, 38, 36, 34, 32, 30, 28, 26, 24, 22, 20, 18, 16, 14, 12, 10, 8, 6, 4)])
        print(sp.boes)

    def test_filter_stocks(self):
        print(f'start test building stock pool function\n')
        ds = QT_DATA_SOURCE
        share_basics = ds.read_table_data('stock_basic')[['symbol', 'name', 'area', 'industry',
                                                          'market', 'list_date', 'exchange']]

        print(f'\nselect all stocks by area')
        stock_pool = qt.filter_stock_codes(area='上海')
        print(f'{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stock areas are "上海"\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['area'].eq('上海').all())

        print(f'\nselect all stocks by multiple areas')
        stock_pool = qt.filter_stock_codes(area='贵州,北京,天津')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stock areas are in list of ["贵州", "北京", "天津"]\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['area'].isin(['贵州',
                                                                                            '北京',
                                                                                            '天津']).all())

        print(f'\nselect all stocks by area and industry')
        stock_pool = qt.filter_stock_codes(area='四川', industry='银行, 金融')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stock areas are "四川", and industry in ["银行", "金融"]\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].head()}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['industry'].isin(['银行', '金融']).all())
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['area'].isin(['四川']).all())

        print(f'\nselect all stocks by industry')
        stock_pool = qt.filter_stock_codes(industry='银行, 金融')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stocks industry in ["银行", "金融"]\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['industry'].isin(['银行', '金融']).all())

        print(f'\nselect all stocks by market')
        stock_pool = qt.filter_stock_codes(market='主板')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stock market is "主板"\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['market'].isin(['主板']).all())

        print(f'\nselect all stocks by market and list date')
        stock_pool = qt.filter_stock_codes(date='2000-01-01', market='主板')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stock market is "主板", and list date after "2000-01-01"\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['market'].isin(['主板']).all())
        date = pd.to_datetime('2000-01-01')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['list_date'].le(date).all())

        print(f'\nselect all stocks by list date')
        stock_pool = qt.filter_stock_codes(date='1997-01-01')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all list date after "1997-01-01"\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        date = pd.to_datetime('1997-01-01')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['list_date'].le(date).all())

        print(f'\nselect all stocks by exchange')
        stock_pool = qt.filter_stock_codes(exchange='SSE')
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all exchanges are in "SSE"\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['exchange'].eq('SSE').all())

        print(f'\nselect all stocks by industry, area and list date')
        industry_list = ['银行', '全国地产', '互联网', '环境保护', '区域地产',
                         '酒店餐饮', '运输设备', '综合类', '建筑工程', '玻璃',
                         '家用电器', '文教休闲', '其他商业', '元器件', 'IT设备',
                         '其他建材', '汽车服务', '火力发电', '医药商业', '汽车配件',
                         '广告包装', '轻工机械', '新型电力', '多元金融', '饲料']
        area_list = ['深圳', '北京', '吉林', '江苏', '辽宁', '广东',
                     '安徽', '四川', '浙江', '湖南', '河北', '新疆',
                     '山东', '河南', '山西', '江西', '青海', '湖北',
                     '内蒙', '海南', '重庆', '陕西', '福建', '广西',
                     '上海']
        stock_pool = qt.filter_stock_codes(date='19980101',
                                           industry=industry_list,
                                           area=area_list)
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all exchanges are in\n{area_list} \nand \n{industry_list}'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        date = pd.to_datetime('1998-01-01')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['list_date'].le(date).all())
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['industry'].isin(industry_list).all())
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['area'].isin(area_list).all())

        self.assertRaises(KeyError, qt.filter_stock_codes, industry=25)
        self.assertRaises(KeyError, qt.filter_stock_codes, share_name='000300.SH')
        self.assertRaises(KeyError, qt.filter_stock_codes, markets='SSE')

        print(f'\nselect all stocks by index, with start and end dates:\n'
              f'all the "000300.SH" composite after 20180101')
        stock_pool = qt.filter_stock_codes(date='20200101',
                                           index='000300.SH')
        self.assertTrue(len(stock_pool) > 0)
        print(f'\n{len(stock_pool)} shares selected, first 10 are: {stock_pool[0:10]}\n'
              f'more information of some fo the stocks\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')

        print(f'\nprint out targets that can not be matched and return fuzzy results')
        stock_pool = qt.filter_stock_codes(industry='银行业, 多元金融, 房地产',
                                           area='陕西省',
                                           market='主要')
        self.assertTrue(len(stock_pool) > 0)
        print(f'\n{len(stock_pool)} shares selected, first 5 are: {stock_pool[0:5]}\n'
              f'check if all stocks industry in ["多元金融"]\n'
              f'{share_basics[np.isin(share_basics.index, stock_pool)].sample(10)}')
        self.assertTrue(share_basics[np.isin(share_basics.index, stock_pool)]['industry'].isin(["多元金融"]).all())

    def test_get_basic_info(self):
        """ 测试获取证券基本信息"""
        qt.get_basic_info('000300')
        qt.get_basic_info('000300.OF')
        qt.get_basic_info('004040')
        qt.get_basic_info('沪镍')
        qt.get_basic_info('中国移动', verbose=True)

    def test_find_history_data(self):
        """ 测试查找打印历史数据信息"""
        qt.find_history_data('pe')
        qt.find_history_data('open')
        qt.find_history_data('市盈率')
        qt.find_history_data('市?率*')
        qt.find_history_data('市值')
        qt.find_history_data('net_profit')
        qt.find_history_data('净利润')

    def test_get_history_data(self):
        """ 测试get_history_data。大部份功能性测试在TestHistoryPanle.test_get_history_panel中完成
        这里只关注参数自动完善功能以及包含完整shares和htypes输入的功能"""
        print('test basic functions')
        print('read data with all parameters and output dataframe grouped by htypes')
        res = qt.get_history_data(shares='000002.SZ, 000001.SZ, 000300.SH',
                                  htypes='open, high, low, close',
                                  start='20210101',
                                  end='20210115',
                                  freq='d',
                                  asset_type='E',
                                  adj='f',
                                  group_by='htypes')
        self.assertIsInstance(res, dict)
        self.assertTrue(all(isinstance(item, pd.DataFrame) for item in res.values()))
        self.assertEqual(list(res.keys()), ['open', 'high', 'low', 'close'])
        first_df = res['open']
        first_index = first_df.index
        first_columns = first_df.columns
        for df in res.values():
            self.assertEqual(list(first_index), list(df.index))
            self.assertEqual(list(first_columns), list(df.columns))

        print('read data with all parameters and output dataframe grouped by shares')
        res = qt.get_history_data(shares='000002.SZ, 000001.SZ, 000300.SH',
                                  htypes='open, high, low, close',
                                  start='20210101',
                                  end='20210115',
                                  freq='d',
                                  asset_type='E',
                                  adj='f',
                                  group_by='shares')
        self.assertIsInstance(res, dict)
        self.assertTrue(all(isinstance(item, pd.DataFrame) for item in res.values()))
        self.assertEqual(list(res.keys()), ['000002.SZ', '000001.SZ', '000300.SH'])
        first_df = res['000002.SZ']
        first_index = first_df.index
        first_columns = first_df.columns
        for df in res.values():
            self.assertEqual(list(first_index), list(df.index))
            self.assertEqual(list(first_columns), list(df.columns))

        print('test function with missing parameters')
        res = qt.get_history_data(htypes='open, high, low, close')
        print(res)

        res = qt.get_history_data(htypes='open, close, vol',
                                  shares='000651.SZ, 513100.SH')
        print(res)

        print('test function with wrong parameters')
        print('wrong share code')
        res = qt.get_history_data(htypes='open, close, vol',
                                  shares='missing_code, 513100.SH')
        print(res)
        print('wrong date')
        self.assertRaises(ValueError,
                          qt.get_history_data,
                          htypes='open, close, vol',
                          shares='missing_code, 513100.SH',
                          start='20220101',
                          end='20210101'
                          )
        self.assertRaises(Exception,
                          qt.get_history_data,
                          htypes='open, close, vol',
                          shares='missing_code, 513100.SH',
                          start='wrong_date',
                          end='20210101'
                          )
        print('wrong freq')
        self.assertRaises(Exception,
                          qt.get_history_data,
                          htypes='open, close, vol',
                          shares='missing_code, 513100.SH',
                          freq='wrong_freq'
                          )
        print('wrong asset_type')
        self.assertRaises(Exception,
                          qt.get_history_data,
                          htypes='open, close, vol',
                          shares='missing_code, 513100.SH',
                          asset_type='wrong_type'
                          )
        print('wrong adj')
        self.assertRaises(Exception,
                          qt.get_history_data,
                          htypes='open, close, vol',
                          shares='missing_code, 513100.SH',
                          adj='wrong_adj'
                          )


if __name__ == '__main__':
    unittest.main()


# coding=utf-8
# ======================================
# File:     test_cost.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Unittest for all qteasy back test
#   cost related functions.
# ======================================
import unittest

import qteasy as qt
import numpy as np

from qteasy.finance import get_selling_result, get_purchase_result, calculate


class TestCost(unittest.TestCase):
    """ 测试所有跟交易成本计算相关的函数"""
    def setUp(self):
        self.amounts = np.array([10000., 20000., 10000.])
        self.op = np.array([0., 1., -0.33333333])
        self.amounts_to_sell = np.array([0., 0., -3333.3333])
        self.cash_to_spend = np.array([0., 20000., 0.])
        self.prices = np.array([10., 20., 10.])
        self.r = qt.set_cost()

    def test_rate_creation(self):
        """测试对象生成"""
        print('testing rates objects\n')
        self.assertIsInstance(self.r, dict, 'Type should be a dictThisis_007')
        self.assertEqual(self.r['buy_fix'], 0)
        self.assertEqual(self.r['sell_fix'], 0)

    def test_rate_operations(self):
        """测试交易费率对象"""
        self.assertEqual(self.r['buy_fix'], 0.0, 'Item got is incorrect')
        self.assertEqual(self.r['sell_fix'], 0.0, 'Item got is wrong')
        self.assertEqual(self.r['buy_rate'], 0.003, 'Item got is incorrect')
        self.assertEqual(self.r['sell_rate'], 0.001, 'Item got is incorrect')
        self.assertEqual(self.r['buy_min'], 5., 'Item got is incorrect')
        self.assertEqual(self.r['sell_min'], 0.0, 'Item got is incorrect')
        self.assertEqual(self.r['slipage'], 0.0, 'Item got is incorrect')
        self.assertEqual(np.allclose(calculate(self.amounts, None, None, **self.r),
                                     [0.003, 0.003, 0.003]),
                         True,
                         'fee calculation wrong')

    def test_rate_fee(self):
        """测试买卖交易费率"""
        self.r['buy_rate'] = 0.003
        self.r['sell_rate'] = 0.001
        self.r['buy_fix'] = 0.
        self.r['sell_fix'] = 0.
        self.r['buy_min'] = 0.
        self.r['sell_min'] = 0.
        self.r['slipage'] = 0.
        r = self.r
        print('\nSell result with fixed rate = 0.001 and moq = 0:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 0, **r))
        test_rate_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 0, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 0., -3333.3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), 33299.999667, msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 33.333333, msg='result incorrect')

        print('\nSell result with fixed rate = 0.001 and moq = 1:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 1., **r))
        test_rate_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 1, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 0., -3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), 33296.67, msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 33.33, msg='result incorrect')

        print('\nSell result with fixed rate = 0.001 and moq = 100:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 100, **r))
        test_rate_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 100, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 0., -3300]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), 32967.0, msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 33, msg='result incorrect')

        print('\nPurchase result with fixed rate = 0.003 and moq = 0:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 0, **r))
        test_rate_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 0, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 997.00897308, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), -20000.0, msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 59.82053838484547, msg='result incorrect')

        print('\nPurchase result with fixed rate = 0.003 and moq = 1:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 1, **r))
        test_rate_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 1, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 997., 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), -19999.82, msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 59.82, msg='result incorrect')

        print('\nPurchase result with fixed rate = 0.003 and moq = 100:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 100, **r))
        test_rate_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 100, **r)
        self.assertIs(np.allclose(test_rate_fee_result[0], [0., 900., 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[1].sum(), -18054., msg='result incorrect')
        self.assertAlmostEqual(test_rate_fee_result[2].sum(), 54.0, msg='result incorrect')

    def test_min_fee(self):
        """测试最低交易费用"""
        self.r['buy_rate'] = 0.
        self.r['sell_rate'] = 0.
        self.r['buy_fix'] = 0.
        self.r['sell_fix'] = 0.
        self.r['buy_min'] = 300
        self.r['sell_min'] = 300
        self.r['slipage'] = 0.
        r = self.r
        print('\npurchase result with fixed cost rate with min fee = 300 and moq = 0:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 0, **r))
        test_min_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 0, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0., 985, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), -20000.0, msg='result incorrect')
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0, msg='result incorrect')

        print('\npurchase result with fixed cost rate with min fee = 300 and moq = 10:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 10, **r))
        test_min_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 10, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0., 980, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), -19900.0, msg='result incorrect')
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0, msg='result incorrect')

        print('\npurchase result with fixed cost rate with min fee = 300 and moq = 100:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 100, **r))
        test_min_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 100, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0., 900, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), -18300.0, msg='result incorrect')
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0, msg='result incorrect')

        print('\nselling result with fixed cost rate with min fee = 300 and moq = 0:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 0, **r))
        test_min_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 0, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0, 0, -3333.3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), 33033.333)
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0)

        print('\nselling result with fixed cost rate with min fee = 300 and moq = 1:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 1, **r))
        test_min_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 1, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0, 0, -3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), 33030)
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0)

        print('\nselling result with fixed cost rate with min fee = 300 and moq = 100:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 100, **r))
        test_min_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 100, **r)
        self.assertIs(np.allclose(test_min_fee_result[0], [0, 0, -3300]), True, 'result incorrect')
        self.assertAlmostEqual(test_min_fee_result[1].sum(), 32700)
        self.assertAlmostEqual(test_min_fee_result[2].sum(), 300.0)

    def test_rate_with_min(self):
        """测试最低交易费用对其他交易费率参数的影响"""
        self.r['buy_rate'] = 0.0153
        self.r['sell_rate'] = 0.01
        self.r['buy_fix'] = 0.
        self.r['sell_fix'] = 0.
        self.r['buy_min'] = 300
        self.r['sell_min'] = 333
        self.r['slipage'] = 0.
        r = self.r
        print('\npurchase result with fixed cost rate with buy_rate = 0.0153, min fee = 300 and moq = 0:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 0, **r))
        test_rate_with_min_result = get_purchase_result(self.prices, self.cash_to_spend, 0, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0., 984.9305624, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), -20000.0, msg='result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 301.3887520929774, msg='result incorrect')

        print('\npurchase result with fixed cost rate with buy_rate = 0.0153, min fee = 300 and moq = 10:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 10, **r))
        test_rate_with_min_result = get_purchase_result(self.prices, self.cash_to_spend, 10, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0., 980, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), -19900.0, msg='result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 300.0, msg='result incorrect')

        print('\npurchase result with fixed cost rate with buy_rate = 0.0153, min fee = 300 and moq = 100:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 100, **r))
        test_rate_with_min_result = get_purchase_result(self.prices, self.cash_to_spend, 100, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0., 900, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), -18300.0, msg='result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 300.0, msg='result incorrect')

        print('\nselling result with fixed cost rate with sell_rate = 0.01, min fee = 333 and moq = 0:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 0, **r))
        test_rate_with_min_result = get_selling_result(self.prices, self.amounts_to_sell, 0, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0, 0, -3333.3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), 32999.99967)
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 333.33333)

        print('\nselling result with fixed cost rate with sell_rate = 0.01, min fee = 333 and moq = 1:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 1, **r))
        test_rate_with_min_result = get_selling_result(self.prices, self.amounts_to_sell, 1, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0, 0, -3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), 32996.7)
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 333.3)

        print('\nselling result with fixed cost rate with sell_rate = 0.01, min fee = 333 and moq = 100:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 100, **r))
        test_rate_with_min_result = get_selling_result(self.prices, self.amounts_to_sell, 100, **r)
        self.assertIs(np.allclose(test_rate_with_min_result[0], [0, 0, -3300]), True, 'result incorrect')
        self.assertAlmostEqual(test_rate_with_min_result[1].sum(), 32667.0)
        self.assertAlmostEqual(test_rate_with_min_result[2].sum(), 333.0)

    def test_fixed_fee(self):
        """测试固定交易费用"""
        self.r['buy_rate'] = 0.
        self.r['sell_rate'] = 0.
        self.r['buy_fix'] = 200
        self.r['sell_fix'] = 150
        self.r['buy_min'] = 0
        self.r['sell_min'] = 0
        self.r['slipage'] = 0
        r = self.r
        print('\nselling result of fixed cost with fixed fee = 150 and moq=0:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 0, **r))
        test_fixed_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 0, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0, 0, -3333.3333]), True, 'result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), 33183.333, msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 150.0, msg='result incorrect')

        print('\nselling result of fixed cost with fixed fee = 150 and moq=100:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 100, **r))
        test_fixed_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 100, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0, 0, -3300.]), True,
                      f'result incorrect, {test_fixed_fee_result[0]} does not equal to [0,0,-3400]')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), 32850., msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 150., msg='result incorrect')

        print('\npurchase result of fixed cost with fixed fee = 200:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 0, **r))
        test_fixed_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 0, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0., 990., 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), -20000.0, msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 200.0, msg='result incorrect')

        print('\npurchase result of fixed cost with fixed fee = 200:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 100, **r))
        test_fixed_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 100, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0., 900., 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), -18200.0, msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 200.0, msg='result incorrect')

    def test_slipage(self):
        """测试交易滑点"""
        self.r['buy_fix'] = 0
        self.r['sell_fix'] = 0
        self.r['buy_min'] = 0
        self.r['sell_min'] = 0
        self.r['buy_rate'] = 0.003
        self.r['sell_rate'] = 0.001
        self.r['slipage'] = 1E-9
        r = self.r
        print('\npurchase result of fixed rate = 0.003 and slipage = 1E-10 and moq = 0:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 0, **r))
        print('\npurchase result of fixed rate = 0.003 and slipage = 1E-10 and moq = 100:')
        print(get_purchase_result(self.prices, self.cash_to_spend, 100, **r))
        print('\nselling result with fixed rate = 0.001 and slipage = 1E-10:')
        print(get_selling_result(self.prices, self.amounts_to_sell, 0, **r))

        test_fixed_fee_result = get_selling_result(self.prices, self.amounts_to_sell, 0, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0, 0, -3333.3333]), True,
                      f'{test_fixed_fee_result[0]} does not equal to [0, 0, -10000]')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), 33298.88855591,
                                msg=f'{test_fixed_fee_result[1]} does not equal to 33298.')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 34.44444409,
                                msg=f'{test_fixed_fee_result[2]} does not equal to -36.666663.')

        test_fixed_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 0, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0., 996.98909294, 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), -20000.0, msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 60.21814121353513, msg='result incorrect')

        test_fixed_fee_result = get_purchase_result(self.prices, self.cash_to_spend, 100, **r)
        self.assertIs(np.allclose(test_fixed_fee_result[0], [0., 900., 0.]), True, 'result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[1].sum(), -18054.36, msg='result incorrect')
        self.assertAlmostEqual(test_fixed_fee_result[2].sum(), 54.36, msg='result incorrect')


if __name__ == '__main__':
    unittest.main()


# ======================================
# File:     setup.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-01-23
# Desc:
#   QTEASY setup file
# ======================================

from setuptools import setup

setup()
